import pytest
import sqlalchemy as sa

from sqlalchemy_utils import Ltree, LtreeType


@pytest.fixture
def Section(Base):
    class Section(Base):
        __tablename__ = 'section'
        id = sa.Column(sa.Integer, primary_key=True)
        path = sa.Column(LtreeType)

    return Section


@pytest.fixture
def init_models(Section, connection):
    connection.execute('CREATE EXTENSION IF NOT EXISTS ltree')
    pass


@pytest.mark.usefixtures('postgresql_dsn')
class TestLTREE(object):
    def test_saves_path(self, session, Section):
        section = Section(path='1.1.2')

        session.add(section)
        session.commit()

        user = session.query(Section).first()
        assert user.path == '1.1.2'

    def test_scalar_attributes_get_coerced_to_objects(self, Section):
        section = Section(path='path.path')
        assert isinstance(section.path, Ltree)

    def test_literal_param(self, session, Section):
        clause = Section.path == 'path'
        compiled = str(clause.compile(compile_kwargs={'literal_binds': True}))
        assert compiled == 'section.path = \'path\''



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import get_referencing_foreign_keys


class TestGetReferencingFksWithCompositeKeys(object):

    @pytest.fixture
    def User(self, Base):
        class User(Base):
            __tablename__ = 'user'
            first_name = sa.Column(sa.Unicode(255), primary_key=True)
            last_name = sa.Column(sa.Unicode(255), primary_key=True)
        return User

    @pytest.fixture
    def Article(self, Base, User):
        class Article(Base):
            __tablename__ = 'article'
            id = sa.Column(sa.Integer, primary_key=True)
            author_first_name = sa.Column(sa.Unicode(255))
            author_last_name = sa.Column(sa.Unicode(255))
            __table_args__ = (
                sa.ForeignKeyConstraint(
                    [author_first_name, author_last_name],
                    [User.first_name, User.last_name]
                ),
            )
        return Article

    @pytest.fixture
    def init_models(self, User, Article):
        pass

    def test_with_declarative_class(self, User, Article):
        fks = get_referencing_foreign_keys(User)
        assert Article.__table__.foreign_keys == fks

    def test_with_table(self, User, Article):
        fks = get_referencing_foreign_keys(User.__table__)
        assert Article.__table__.foreign_keys == fks


class TestGetReferencingFksWithInheritance(object):

    @pytest.fixture
    def User(self, Base):
        class User(Base):
            __tablename__ = 'user'
            id = sa.Column(sa.Integer, primary_key=True)
            type = sa.Column(sa.Unicode)
            first_name = sa.Column(sa.Unicode(255))
            last_name = sa.Column(sa.Unicode(255))

            __mapper_args__ = {
                'polymorphic_on': 'type'
            }
        return User

    @pytest.fixture
    def Admin(self, User):
        class Admin(User):
            __tablename__ = 'admin'
            id = sa.Column(
                sa.Integer, sa.ForeignKey(User.id), primary_key=True
            )
        return Admin

    @pytest.fixture
    def TextItem(self, Base, User):
        class TextItem(Base):
            __tablename__ = 'textitem'
            id = sa.Column(sa.Integer, primary_key=True)
            type = sa.Column(sa.Unicode)
            author_id = sa.Column(sa.Integer, sa.ForeignKey(User.id))
            __mapper_args__ = {
                'polymorphic_on': 'type'
            }
        return TextItem

    @pytest.fixture
    def Article(self, TextItem):
        class Article(TextItem):
            __tablename__ = 'article'
            id = sa.Column(
                sa.Integer, sa.ForeignKey(TextItem.id), primary_key=True
            )
            __mapper_args__ = {
                'polymorphic_identity': 'article'
            }
        return Article

    @pytest.fixture
    def init_models(self, User, Admin, TextItem, Article):
        pass

    def test_with_declarative_class(self, Admin, TextItem):
        fks = get_referencing_foreign_keys(Admin)
        assert TextItem.__table__.foreign_keys == fks

    def test_with_table(self, Admin):
        fks = get_referencing_foreign_keys(Admin.__table__)
        assert fks == set([])



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import get_tables


@pytest.fixture
def TextItem(Base):
    class TextItem(Base):
        __tablename__ = 'text_item'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))
        type = sa.Column(sa.Unicode(255))

        __mapper_args__ = {
            'polymorphic_on': type,
            'with_polymorphic': '*'
        }
    return TextItem


@pytest.fixture
def Article(TextItem):
    class Article(TextItem):
        __tablename__ = 'article'
        id = sa.Column(
            sa.Integer, sa.ForeignKey(TextItem.id), primary_key=True
        )
        __mapper_args__ = {
            'polymorphic_identity': u'article'
        }
    return Article


@pytest.fixture
def init_models(TextItem, Article):
    pass


class TestGetTables(object):

    def test_child_class_using_join_table_inheritance(self, TextItem, Article):
        assert get_tables(Article) == [
            TextItem.__table__,
            Article.__table__
        ]

    def test_entity_using_with_polymorphic(self, TextItem, Article):
        assert get_tables(TextItem) == [
            TextItem.__table__,
            Article.__table__
        ]

    def test_instrumented_attribute(self, TextItem):
        assert get_tables(TextItem.name) == [
            TextItem.__table__,
        ]

    def test_polymorphic_instrumented_attribute(self, TextItem, Article):
        assert get_tables(Article.id) == [
            TextItem.__table__,
            Article.__table__
        ]

    def test_column(self, Article):
        assert get_tables(Article.__table__.c.id) == [
            Article.__table__
        ]

    def test_mapper_entity_with_class(self, session, TextItem, Article):
        query = session.query(Article)
        assert get_tables(query._entities[0]) == [
            TextItem.__table__, Article.__table__
        ]

    def test_mapper_entity_with_mapper(self, session, TextItem, Article):
        query = session.query(sa.inspect(Article))
        assert get_tables(query._entities[0]) == [
            TextItem.__table__, Article.__table__
        ]

    def test_column_entity(self, session, TextItem, Article):
        query = session.query(Article.id)
        assert get_tables(query._entities[0]) == [
            TextItem.__table__, Article.__table__
        ]



import pytest
import pytz
import sqlalchemy as sa
from dateutil.zoneinfo import getzoneinfofile_stream, tzfile, ZoneInfoFile

from sqlalchemy_utils.types import timezone, TimezoneType


@pytest.fixture
def Visitor(Base):
    class Visitor(Base):
        __tablename__ = 'visitor'
        id = sa.Column(sa.Integer, primary_key=True)
        timezone_dateutil = sa.Column(
            timezone.TimezoneType(backend='dateutil')
        )
        timezone_pytz = sa.Column(
            timezone.TimezoneType(backend='pytz')
        )

        def __repr__(self):
            return 'Visitor(%r)' % self.id
    return Visitor


@pytest.fixture
def init_models(Visitor):
    pass


class TestTimezoneType(object):

    def test_parameter_processing(self, session, Visitor):
        visitor = Visitor(
            timezone_dateutil=u'America/Los_Angeles',
            timezone_pytz=u'America/Los_Angeles'
        )

        session.add(visitor)
        session.commit()

        visitor_dateutil = session.query(Visitor).filter_by(
            timezone_dateutil=u'America/Los_Angeles'
        ).first()
        visitor_pytz = session.query(Visitor).filter_by(
            timezone_pytz=u'America/Los_Angeles'
        ).first()

        assert visitor_dateutil is not None
        assert visitor_pytz is not None


TIMEZONE_BACKENDS = ['dateutil', 'pytz']


def test_can_coerce_pytz_DstTzInfo():
    tzcol = TimezoneType(backend='pytz')
    tz = pytz.timezone('America/New_York')
    assert isinstance(tz, pytz.tzfile.DstTzInfo)
    assert tzcol._coerce(tz) is tz


def test_can_coerce_pytz_StaticTzInfo():
    tzcol = TimezoneType(backend='pytz')
    tz = pytz.timezone('Pacific/Truk')
    assert tzcol._coerce(tz) is tz


@pytest.mark.parametrize('zone', pytz.all_timezones)
def test_can_coerce_string_for_pytz_zone(zone):
    tzcol = TimezoneType(backend='pytz')
    assert tzcol._coerce(zone).zone == zone


@pytest.mark.parametrize(
    'zone', ZoneInfoFile(getzoneinfofile_stream()).zones.keys())
def test_can_coerce_string_for_dateutil_zone(zone):
    tzcol = TimezoneType(backend='dateutil')
    assert isinstance(tzcol._coerce(zone), tzfile)


@pytest.mark.parametrize('backend', TIMEZONE_BACKENDS)
def test_can_coerce_and_raise_UnknownTimeZoneError_or_ValueError(backend):
    tzcol = TimezoneType(backend=backend)
    with pytest.raises((ValueError, pytz.exceptions.UnknownTimeZoneError)):
        tzcol._coerce('SolarSystem/Mars')
    with pytest.raises((ValueError, pytz.exceptions.UnknownTimeZoneError)):
        tzcol._coerce('')


@pytest.mark.parametrize('backend', TIMEZONE_BACKENDS)
def test_can_coerce_None(backend):
    tzcol = TimezoneType(backend=backend)
    assert tzcol._coerce(None) is None



import pytest
import sqlalchemy as sa
from sqlalchemy.ext.declarative import declarative_base

from sqlalchemy_utils import cast_if


@pytest.fixture(scope='class')
def base():
    return declarative_base()


@pytest.fixture(scope='class')
def article_cls(base):
    class Article(base):
        __tablename__ = 'article'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.String)
        name_synonym = sa.orm.synonym('name')

    return Article


class TestCastIf(object):
    def test_column(self, article_cls):
        expr = article_cls.__table__.c.name
        assert cast_if(expr, sa.String) is expr

    def test_column_property(self, article_cls):
        expr = article_cls.name.property
        assert cast_if(expr, sa.String) is expr

    def test_instrumented_attribute(self, article_cls):
        expr = article_cls.name
        assert cast_if(expr, sa.String) is expr

    def test_synonym(self, article_cls):
        expr = article_cls.name_synonym
        assert cast_if(expr, sa.String) is expr

    def test_scalar_selectable(self, article_cls):
        expr = sa.select([article_cls.id]).as_scalar()
        assert cast_if(expr, sa.Integer) is expr

    def test_scalar(self):
        assert cast_if('something', sa.String) == 'something'



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import get_fk_constraint_for_columns, has_unique_index


class TestHasUniqueIndex(object):

    @pytest.fixture
    def articles(self, Base):
        class Article(Base):
            __tablename__ = 'article'
            id = sa.Column(sa.Integer, primary_key=True)
        return Article.__table__

    @pytest.fixture
    def article_translations(self, Base):
        class ArticleTranslation(Base):
            __tablename__ = 'article_translation'
            id = sa.Column(sa.Integer, primary_key=True)
            locale = sa.Column(sa.String(10), primary_key=True)
            title = sa.Column(sa.String(100))
            is_published = sa.Column(sa.Boolean, index=True)
            is_deleted = sa.Column(sa.Boolean, unique=True)
            is_archived = sa.Column(sa.Boolean)

            __table_args__ = (
                sa.Index('my_index', is_archived, is_published, unique=True),
            )

        return ArticleTranslation.__table__

    def test_primary_key(self, articles):
        assert has_unique_index(articles.c.id)

    def test_column_of_aliased_table(self, articles):
        alias = sa.orm.aliased(articles)
        with pytest.raises(TypeError):
            assert has_unique_index(alias.c.id)

    def test_unique_index(self, article_translations):
        assert has_unique_index(article_translations.c.is_deleted)

    def test_compound_primary_key(self, article_translations):
        assert not has_unique_index(article_translations.c.id)
        assert not has_unique_index(article_translations.c.locale)

    def test_single_column_index(self, article_translations):
        assert not has_unique_index(article_translations.c.is_published)

    def test_compound_column_unique_index(self, article_translations):
        assert not has_unique_index(article_translations.c.is_published)
        assert not has_unique_index(article_translations.c.is_archived)


class TestHasUniqueIndexWithFKConstraint(object):
    def test_composite_fk_without_index(self, Base):

        class User(Base):
            __tablename__ = 'user'
            first_name = sa.Column(sa.Unicode(255), primary_key=True)
            last_name = sa.Column(sa.Unicode(255), primary_key=True)

        class Article(Base):
            __tablename__ = 'article'
            id = sa.Column(sa.Integer, primary_key=True)
            author_first_name = sa.Column(sa.Unicode(255))
            author_last_name = sa.Column(sa.Unicode(255))
            __table_args__ = (
                sa.ForeignKeyConstraint(
                    [author_first_name, author_last_name],
                    [User.first_name, User.last_name]
                ),
            )

        table = Article.__table__
        constraint = get_fk_constraint_for_columns(
            table,
            table.c.author_first_name,
            table.c.author_last_name
        )
        assert not has_unique_index(constraint)

    def test_composite_fk_with_index(self, Base):

        class User(Base):
            __tablename__ = 'user'
            first_name = sa.Column(sa.Unicode(255), primary_key=True)
            last_name = sa.Column(sa.Unicode(255), primary_key=True)

        class Article(Base):
            __tablename__ = 'article'
            id = sa.Column(sa.Integer, primary_key=True)
            author_first_name = sa.Column(sa.Unicode(255))
            author_last_name = sa.Column(sa.Unicode(255))
            __table_args__ = (
                sa.ForeignKeyConstraint(
                    [author_first_name, author_last_name],
                    [User.first_name, User.last_name]
                ),
                sa.Index(
                    'my_index',
                    author_first_name,
                    author_last_name,
                    unique=True
                )
            )

        table = Article.__table__
        constraint = get_fk_constraint_for_columns(
            table,
            table.c.author_first_name,
            table.c.author_last_name
        )
        assert has_unique_index(constraint)

    def test_composite_fk_with_partial_index_match(self, Base):

        class User(Base):
            __tablename__ = 'user'
            first_name = sa.Column(sa.Unicode(255), primary_key=True)
            last_name = sa.Column(sa.Unicode(255), primary_key=True)

        class Article(Base):
            __tablename__ = 'article'
            id = sa.Column(sa.Integer, primary_key=True)
            author_first_name = sa.Column(sa.Unicode(255))
            author_last_name = sa.Column(sa.Unicode(255))
            __table_args__ = (
                sa.ForeignKeyConstraint(
                    [author_first_name, author_last_name],
                    [User.first_name, User.last_name]
                ),
                sa.Index(
                    'my_index',
                    author_first_name,
                    author_last_name,
                    id,
                    unique=True
                )
            )

        table = Article.__table__
        constraint = get_fk_constraint_for_columns(
            table,
            table.c.author_first_name,
            table.c.author_last_name
        )
        assert not has_unique_index(constraint)



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import aggregated, TSVectorType


def tsvector_reduce_concat(vectors):
    return sa.sql.expression.cast(
        sa.func.coalesce(
            sa.func.array_to_string(sa.func.array_agg(vectors), ' ')
        ),
        TSVectorType
    )


@pytest.fixture
def Product(Base):
    class Product(Base):
        __tablename__ = 'product'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))
        price = sa.Column(sa.Numeric)

        catalog_id = sa.Column(sa.Integer, sa.ForeignKey('catalog.id'))
    return Product


@pytest.fixture
def Catalog(Base, Product):
    class Catalog(Base):
        __tablename__ = 'catalog'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))

        @aggregated('products', sa.Column(TSVectorType))
        def product_search_vector(self):
            return tsvector_reduce_concat(
                sa.func.to_tsvector(Product.name)
            )

        products = sa.orm.relationship('Product', backref='catalog')
    return Catalog


@pytest.fixture
def init_models(Product, Catalog):
    pass


@pytest.mark.usefixtures('postgresql_dsn')
class TestSearchVectorAggregates(object):

    def test_assigns_aggregates_on_insert(self, session, Product, Catalog):
        catalog = Catalog(
            name=u'Some catalog'
        )
        session.add(catalog)
        session.commit()
        product = Product(
            name=u'Product XYZ',
            catalog=catalog
        )
        session.add(product)
        session.commit()
        session.refresh(catalog)
        assert catalog.product_search_vector == "'product':1 'xyz':2"



import pytest
import sqlalchemy as sa

from sqlalchemy_utils.functions import getdotattr


@pytest.fixture
def Document(Base):
    class Document(Base):
        __tablename__ = 'document'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))
    return Document


@pytest.fixture
def Section(Base, Document):
    class Section(Base):
        __tablename__ = 'section'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))

        document_id = sa.Column(
            sa.Integer, sa.ForeignKey(Document.id)
        )

        document = sa.orm.relationship(Document, backref='sections')
    return Section


@pytest.fixture
def SubSection(Base, Section):
    class SubSection(Base):
        __tablename__ = 'subsection'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))

        section_id = sa.Column(
            sa.Integer, sa.ForeignKey(Section.id)
        )

        section = sa.orm.relationship(Section, backref='subsections')
    return SubSection


@pytest.fixture
def SubSubSection(Base, SubSection):
    class SubSubSection(Base):
        __tablename__ = 'subsubsection'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))
        locale = sa.Column(sa.String(10))

        subsection_id = sa.Column(
            sa.Integer, sa.ForeignKey(SubSection.id)
        )

        subsection = sa.orm.relationship(
            SubSection, backref='subsubsections'
        )
    return SubSubSection


@pytest.fixture
def init_models(Document, Section, SubSection, SubSubSection):
    pass


class TestGetDotAttr(object):

    def test_simple_objects(self, Document, Section, SubSection):
        document = Document(name=u'some document')
        section = Section(document=document)
        subsection = SubSection(section=section)

        assert getdotattr(
            subsection,
            'section.document.name'
        ) == u'some document'

    def test_with_instrumented_lists(
        self,
        Document,
        Section,
        SubSection,
        SubSubSection
    ):
        document = Document(name=u'some document')
        section = Section(document=document)
        subsection = SubSection(section=section)
        subsubsection = SubSubSection(subsection=subsection)

        assert getdotattr(document, 'sections') == [section]
        assert getdotattr(document, 'sections.subsections') == [
            subsection
        ]
        assert getdotattr(document, 'sections.subsections.subsubsections') == [
            subsubsection
        ]

    def test_class_paths(self, Document, Section, SubSection):
        assert getdotattr(Section, 'document') is Section.document
        assert (
            getdotattr(SubSection, 'section.document') is
            Section.document
        )
        assert getdotattr(Section, 'document.name') is Document.name



from sqlalchemy.dialects import postgresql

from sqlalchemy_utils.functions import quote


class TestQuote(object):
    def test_quote_with_preserved_keyword(self, engine, connection, session):
        assert quote(connection, 'order') == '"order"'
        assert quote(session, 'order') == '"order"'
        assert quote(engine, 'order') == '"order"'
        assert quote(postgresql.dialect(), 'order') == '"order"'

    def test_quote_with_non_preserved_keyword(
        self,
        engine,
        connection,
        session
    ):
        assert quote(connection, 'some_order') == 'some_order'
        assert quote(session, 'some_order') == 'some_order'
        assert quote(engine, 'some_order') == 'some_order'
        assert quote(postgresql.dialect(), 'some_order') == 'some_order'



import pytest
import six
import sqlalchemy as sa
from sqlalchemy.util.langhelpers import symbol

from sqlalchemy_utils.path import AttrPath, Path


@pytest.fixture
def Document(Base):
    class Document(Base):
        __tablename__ = 'document'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))
        locale = sa.Column(sa.String(10))
    return Document


@pytest.fixture
def Section(Base, Document):
    class Section(Base):
        __tablename__ = 'section'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))
        locale = sa.Column(sa.String(10))

        document_id = sa.Column(
            sa.Integer, sa.ForeignKey(Document.id)
        )

        document = sa.orm.relationship(Document, backref='sections')
    return Section


@pytest.fixture
def SubSection(Base, Section):
    class SubSection(Base):
        __tablename__ = 'subsection'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))
        locale = sa.Column(sa.String(10))

        section_id = sa.Column(
            sa.Integer, sa.ForeignKey(Section.id)
        )

        section = sa.orm.relationship(Section, backref='subsections')
    return SubSection


class TestAttrPath(object):

    @pytest.fixture
    def init_models(self, Document, Section, SubSection):
        pass

    def test_direction(self, SubSection):
        assert (
            AttrPath(SubSection, 'section').direction == symbol('MANYTOONE')
        )

    def test_invert(self, Document, Section, SubSection):
        path = ~ AttrPath(SubSection, 'section.document')
        assert path.parts == [
            Document.sections,
            Section.subsections
        ]
        assert str(path.path) == 'sections.subsections'

    def test_len(self, SubSection):
        len(AttrPath(SubSection, 'section.document')) == 2

    def test_init(self, SubSection):
        path = AttrPath(SubSection, 'section.document')
        assert path.class_ == SubSection
        assert path.path == Path('section.document')

    def test_iter(self, Section, SubSection):
        path = AttrPath(SubSection, 'section.document')
        assert list(path) == [
            SubSection.section,
            Section.document
        ]

    def test_repr(self, SubSection):
        path = AttrPath(SubSection, 'section.document')
        assert repr(path) == (
            "AttrPath(SubSection, 'section.document')"
        )

    def test_index(self, Section, SubSection):
        path = AttrPath(SubSection, 'section.document')
        assert path.index(Section.document) == 1
        assert path.index(SubSection.section) == 0

    def test_getitem(self, Section, SubSection):
        path = AttrPath(SubSection, 'section.document')
        assert path[0] is SubSection.section
        assert path[1] is Section.document

    def test_getitem_with_slice(self, Section, SubSection):
        path = AttrPath(SubSection, 'section.document')
        assert path[:] == AttrPath(SubSection, 'section.document')
        assert path[:-1] == AttrPath(SubSection, 'section')
        assert path[1:] == AttrPath(Section, 'document')

    def test_eq(self, SubSection):
        assert (
            AttrPath(SubSection, 'section.document') ==
            AttrPath(SubSection, 'section.document')
        )
        assert not (
            AttrPath(SubSection, 'section') ==
            AttrPath(SubSection, 'section.document')
        )

    def test_ne(self, SubSection):
        assert not (
            AttrPath(SubSection, 'section.document') !=
            AttrPath(SubSection, 'section.document')
        )
        assert (
            AttrPath(SubSection, 'section') !=
            AttrPath(SubSection, 'section.document')
        )


class TestPath(object):
    def test_init(self):
        path = Path('attr.attr2')
        assert path.path == 'attr.attr2'

    def test_init_with_path_object(self):
        path = Path(Path('attr.attr2'))
        assert path.path == 'attr.attr2'

    def test_iter(self):
        path = Path('s.s2.s3')
        assert list(path) == ['s', 's2', 's3']

    @pytest.mark.parametrize(('path', 'length'), (
        (Path('s.s2.s3'), 3),
        (Path('s.s2'), 2),
        (Path(''), 0)
    ))
    def test_len(self, path, length):
        return len(path) == length

    def test_reversed(self):
        path = Path('s.s2.s3')
        assert list(reversed(path)) == ['s3', 's2', 's']

    def test_repr(self):
        path = Path('s.s2')
        assert repr(path) == "Path('s.s2')"

    def test_getitem(self):
        path = Path('s.s2')
        assert path[0] == 's'
        assert path[1] == 's2'

    def test_str(self):
        assert str(Path('s.s2')) == 's.s2'

    def test_index(self):
        assert Path('s.s2.s3').index('s2') == 1

    def test_unicode(self):
        assert six.text_type(Path('s.s2')) == u's.s2'

    def test_getitem_with_slice(self):
        path = Path('s.s2.s3')
        assert path[1:] == Path('s2.s3')

    @pytest.mark.parametrize(('test', 'result'), (
        (Path('s.s2') == Path('s.s2'), True),
        (Path('s.s2') == Path('s.s3'), False)
    ))
    def test_eq(self, test, result):
        assert test is result

    @pytest.mark.parametrize(('test', 'result'), (
        (Path('s.s2') != Path('s.s2'), False),
        (Path('s.s2') != Path('s.s3'), True)
    ))
    def test_ne(self, test, result):
        assert test is result



import pytest
import sqlalchemy as sa
from flexmock import flexmock

from sqlalchemy_utils import Choice, ChoiceType, ImproperlyConfigured
from sqlalchemy_utils.types.choice import Enum


class TestChoice(object):
    def test_equality_operator(self):
        assert Choice(1, 1) == 1
        assert 1 == Choice(1, 1)
        assert Choice(1, 1) == Choice(1, 1)

    def test_non_equality_operator(self):
        assert Choice(1, 1) != 2
        assert not (Choice(1, 1) != 1)

    def test_hash(self):
        assert hash(Choice(1, 1)) == hash(1)


class TestChoiceType(object):

    @pytest.fixture
    def User(self, Base):

        class User(Base):
            TYPES = [
                ('admin', 'Admin'),
                ('regular-user', 'Regular user')
            ]

            __tablename__ = 'user'
            id = sa.Column(sa.Integer, primary_key=True)
            type = sa.Column(ChoiceType(TYPES))

            def __repr__(self):
                return 'User(%r)' % self.id

        return User

    @pytest.fixture
    def init_models(self, User):
        pass

    def test_python_type(self, User):
        type_ = User.__table__.c.type.type
        assert type_.python_type

    def test_string_processing(self, session, User):
        flexmock(ChoiceType).should_receive('_coerce').and_return(
            u'admin'
        )
        user = User(
            type=u'admin'
        )

        session.add(user)
        session.commit()

        user = session.query(User).first()
        assert user.type.value == u'Admin'

    def test_parameter_processing(self, session, User):
        user = User(
            type=u'admin'
        )

        session.add(user)
        session.commit()

        user = session.query(User).first()
        assert user.type.value == u'Admin'

    def test_scalar_attributes_get_coerced_to_objects(self, User):
        user = User(type=u'admin')

        assert isinstance(user.type, Choice)

    def test_throws_exception_if_no_choices_given(self):
        with pytest.raises(ImproperlyConfigured):
            ChoiceType([])


class TestChoiceTypeWithCustomUnderlyingType(object):
    def test_init_type(self):
        type_ = ChoiceType([(1, u'something')], impl=sa.Integer)
        assert type_.impl == sa.Integer


@pytest.mark.skipif('Enum is None')
class TestEnumType(object):

    @pytest.fixture
    def OrderStatus(self):
        class OrderStatus(Enum):
            unpaid = 0
            paid = 1
        return OrderStatus

    @pytest.fixture
    def Order(self, Base, OrderStatus):

        class Order(Base):
            __tablename__ = 'order'
            id_ = sa.Column(sa.Integer, primary_key=True)
            status = sa.Column(
                ChoiceType(OrderStatus, impl=sa.Integer()),
                default=OrderStatus.unpaid,
            )

            def __repr__(self):
                return 'Order(%r, %r)' % (self.id_, self.status)

        return Order

    @pytest.fixture
    def OrderNullable(self, Base, OrderStatus):

        class OrderNullable(Base):
            __tablename__ = 'order_nullable'
            id_ = sa.Column(sa.Integer, primary_key=True)
            status = sa.Column(
                ChoiceType(OrderStatus, impl=sa.Integer()),
                nullable=True,
            )

        return OrderNullable

    @pytest.fixture
    def init_models(self, Order, OrderNullable):
        pass

    def test_parameter_initialization(self, session, Order, OrderStatus):
        order = Order()

        session.add(order)
        session.commit()

        order = session.query(Order).first()
        assert order.status is OrderStatus.unpaid
        assert order.status.value == 0

    def test_setting_by_value(self, session, Order, OrderStatus):
        order = Order()
        order.status = 1

        session.add(order)
        session.commit()

        order = session.query(Order).first()
        assert order.status is OrderStatus.paid

    def test_setting_by_enum(self, session, Order, OrderStatus):
        order = Order()
        order.status = OrderStatus.paid

        session.add(order)
        session.commit()

        order = session.query(Order).first()
        assert order.status is OrderStatus.paid

    def test_setting_value_that_resolves_to_none(
        self,
        session,
        Order,
        OrderStatus
    ):
        order = Order()
        order.status = 0

        session.add(order)
        session.commit()

        order = session.query(Order).first()
        assert order.status is OrderStatus.unpaid

    def test_setting_to_wrong_enum_raises_valueerror(self, Order):
        class WrongEnum(Enum):
            foo = 0
            bar = 1

        order = Order()

        with pytest.raises(ValueError):
            order.status = WrongEnum.foo

    def test_setting_to_uncoerceable_type_raises_valueerror(self, Order):
        order = Order()
        with pytest.raises(ValueError):
            order.status = 'Bad value'

    def test_order_nullable_stores_none(self, session, OrderNullable):
        # With nullable=False as in `Order`, a `None` value is always
        # converted to the default value, unless we explicitly set it to
        # sqlalchemy.sql.null(), so we use this class to test our ability
        # to set and retrive `None`.
        order_nullable = OrderNullable()
        assert order_nullable.status is None

        order_nullable.status = None

        session.add(order_nullable)
        session.commit()

        assert order_nullable.status is None



import pytest
import sqlalchemy as sa

from sqlalchemy_utils.functions import identity


class IdentityTestCase(object):

    @pytest.fixture
    def init_models(self, Building):
        pass

    def test_for_transient_class_without_id(self, Building):
        assert identity(Building()) == (None, )

    def test_for_transient_class_with_id(self, session, Building):
        building = Building(name=u'Some building')
        session.add(building)
        session.flush()

        assert identity(building) == (building.id, )

    def test_identity_for_class(self, Building):
        assert identity(Building) == (Building.id, )


class TestIdentity(IdentityTestCase):

    @pytest.fixture
    def Building(self, Base):
        class Building(Base):
            __tablename__ = 'building'
            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.Unicode(255))
        return Building


class TestIdentityWithColumnAlias(IdentityTestCase):

    @pytest.fixture
    def Building(self, Base):
        class Building(Base):
            __tablename__ = 'building'
            id = sa.Column('_id', sa.Integer, primary_key=True)
            name = sa.Column(sa.Unicode(255))
        return Building



from decimal import Decimal

import pytest
import sqlalchemy as sa

from sqlalchemy_utils.aggregates import aggregated


@pytest.fixture
def Product(Base):
    class Product(Base):
        __tablename__ = 'product'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))
        price = sa.Column(sa.Numeric)
        type = sa.Column(sa.String(255))

        __mapper_args__ = {
            'polymorphic_on': type,
            'polymorphic_identity': 'simple'
        }

        catalog_id = sa.Column(sa.Integer, sa.ForeignKey('catalog.id'))
    return Product


@pytest.fixture(params=['simple', 'child'])
def AnyProduct(Product, request):
    if request.param == 'simple':
        return Product

    class ChildProduct(Product):
        __tablename__ = 'child_product'
        id = sa.Column(
            sa.Integer, sa.ForeignKey(Product.id), primary_key=True
        )

        __mapper_args__ = {
            'polymorphic_identity': 'child',
        }
    return ChildProduct


@pytest.fixture
def Catalog(Base, Product):
    class Catalog(Base):
        __tablename__ = 'catalog'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))
        type = sa.Column(sa.Unicode(255))

        __mapper_args__ = {
            'polymorphic_on': type
        }

        @aggregated('products', sa.Column(sa.Numeric, default=0))
        def net_worth(self):
            return sa.func.sum(Product.price)

        products = sa.orm.relationship('Product', backref='catalog')
    return Catalog


@pytest.fixture
def CostumeCatalog(Catalog):
    class CostumeCatalog(Catalog):
        __tablename__ = 'costume_catalog'
        id = sa.Column(
            sa.Integer, sa.ForeignKey(Catalog.id), primary_key=True
        )

        __mapper_args__ = {
            'polymorphic_identity': 'costumes',
        }
    return CostumeCatalog


@pytest.fixture
def CarCatalog(Catalog):
    class CarCatalog(Catalog):
        __tablename__ = 'car_catalog'
        id = sa.Column(
            sa.Integer, sa.ForeignKey(Catalog.id), primary_key=True
        )

        __mapper_args__ = {
            'polymorphic_identity': 'cars',
        }
    return CarCatalog


@pytest.fixture
def init_models(AnyProduct, Catalog, CostumeCatalog, CarCatalog):
    pass


@pytest.mark.usefixtures('postgresql_dsn')
class TestLazyEvaluatedSelectExpressionsForAggregates(object):

    def test_columns_inherited_from_parent(
        self,
        Catalog,
        CarCatalog,
        CostumeCatalog
    ):
        assert CarCatalog.net_worth
        assert CostumeCatalog.net_worth
        assert Catalog.net_worth
        assert not hasattr(CarCatalog.__table__.c, 'net_worth')
        assert not hasattr(CostumeCatalog.__table__.c, 'net_worth')

    def test_assigns_aggregates_on_insert(self, session, AnyProduct, Catalog):
        catalog = Catalog(
            name=u'Some catalog'
        )
        session.add(catalog)
        session.commit()
        product = AnyProduct(
            name=u'Some product',
            price=Decimal('1000'),
            catalog=catalog
        )
        session.add(product)
        session.commit()
        session.refresh(catalog)
        assert catalog.net_worth == Decimal('1000')

    def test_assigns_aggregates_on_update(self, session, Catalog, AnyProduct):
        catalog = Catalog(
            name=u'Some catalog'
        )
        session.add(catalog)
        session.commit()
        product = AnyProduct(
            name=u'Some product',
            price=Decimal('1000'),
            catalog=catalog
        )
        session.add(product)
        session.commit()
        product.price = Decimal('500')
        session.commit()
        session.refresh(catalog)
        assert catalog.net_worth == Decimal('500')



from sqlalchemy_utils.functions import naturally_equivalent


class TestNaturallyEquivalent(object):
    def test_returns_true_when_properties_match(self, User):
        assert naturally_equivalent(
            User(name=u'someone'), User(name=u'someone')
        )

    def test_skips_primary_keys(self, User):
        assert naturally_equivalent(
            User(id=1, name=u'someone'), User(id=2, name=u'someone')
        )



import pytest
import sqlalchemy as sa


class ThreeLevelDeepOneToOne(object):

    @pytest.fixture
    def Catalog(self, Base, Category):
        class Catalog(Base):
            __tablename__ = 'catalog'
            id = sa.Column('_id', sa.Integer, primary_key=True)
            category = sa.orm.relationship(
                Category,
                uselist=False,
                backref='catalog'
            )
        return Catalog

    @pytest.fixture
    def Category(self, Base, SubCategory):
        class Category(Base):
            __tablename__ = 'category'
            id = sa.Column('_id', sa.Integer, primary_key=True)
            catalog_id = sa.Column(
                '_catalog_id',
                sa.Integer,
                sa.ForeignKey('catalog._id')
            )

            sub_category = sa.orm.relationship(
                SubCategory,
                uselist=False,
                backref='category'
            )
        return Category

    @pytest.fixture
    def SubCategory(self, Base, Product):
        class SubCategory(Base):
            __tablename__ = 'sub_category'
            id = sa.Column('_id', sa.Integer, primary_key=True)
            category_id = sa.Column(
                '_category_id',
                sa.Integer,
                sa.ForeignKey('category._id')
            )
            product = sa.orm.relationship(
                Product,
                uselist=False,
                backref='sub_category'
            )
        return SubCategory

    @pytest.fixture
    def Product(self, Base):
        class Product(Base):
            __tablename__ = 'product'
            id = sa.Column('_id', sa.Integer, primary_key=True)
            price = sa.Column(sa.Integer)

            sub_category_id = sa.Column(
                '_sub_category_id',
                sa.Integer,
                sa.ForeignKey('sub_category._id')
            )
        return Product

    @pytest.fixture
    def init_models(self, Catalog, Category, SubCategory, Product):
        pass


class ThreeLevelDeepOneToMany(object):

    @pytest.fixture
    def Catalog(self, Base, Category):
        class Catalog(Base):
            __tablename__ = 'catalog'
            id = sa.Column('_id', sa.Integer, primary_key=True)

            categories = sa.orm.relationship(Category, backref='catalog')
        return Catalog

    @pytest.fixture
    def Category(self, Base, SubCategory):
        class Category(Base):
            __tablename__ = 'category'
            id = sa.Column('_id', sa.Integer, primary_key=True)
            catalog_id = sa.Column(
                '_catalog_id',
                sa.Integer,
                sa.ForeignKey('catalog._id')
            )

            sub_categories = sa.orm.relationship(
                SubCategory, backref='category'
            )
        return Category

    @pytest.fixture
    def SubCategory(self, Base, Product):
        class SubCategory(Base):
            __tablename__ = 'sub_category'
            id = sa.Column('_id', sa.Integer, primary_key=True)
            category_id = sa.Column(
                '_category_id',
                sa.Integer,
                sa.ForeignKey('category._id')
            )
            products = sa.orm.relationship(
                Product,
                backref='sub_category'
            )
        return SubCategory

    @pytest.fixture
    def Product(self, Base):
        class Product(Base):
            __tablename__ = 'product'
            id = sa.Column('_id', sa.Integer, primary_key=True)
            price = sa.Column(sa.Numeric)

            sub_category_id = sa.Column(
                '_sub_category_id',
                sa.Integer,
                sa.ForeignKey('sub_category._id')
            )

            def __repr__(self):
                return '<Product id=%r>' % self.id
        return Product

    @pytest.fixture
    def init_models(self, Catalog, Category, SubCategory, Product):
        pass


class ThreeLevelDeepManyToMany(object):

    @pytest.fixture
    def Catalog(self, Base, Category):

        catalog_category = sa.Table(
            'catalog_category',
            Base.metadata,
            sa.Column('catalog_id', sa.Integer, sa.ForeignKey('catalog._id')),
            sa.Column('category_id', sa.Integer, sa.ForeignKey('category._id'))
        )

        class Catalog(Base):
            __tablename__ = 'catalog'
            id = sa.Column('_id', sa.Integer, primary_key=True)

            categories = sa.orm.relationship(
                Category,
                backref='catalogs',
                secondary=catalog_category
            )
        return Catalog

    @pytest.fixture
    def Category(self, Base, SubCategory):

        category_subcategory = sa.Table(
            'category_subcategory',
            Base.metadata,
            sa.Column(
                'category_id',
                sa.Integer,
                sa.ForeignKey('category._id')
            ),
            sa.Column(
                'subcategory_id',
                sa.Integer,
                sa.ForeignKey('sub_category._id')
            )
        )

        class Category(Base):
            __tablename__ = 'category'
            id = sa.Column('_id', sa.Integer, primary_key=True)

            sub_categories = sa.orm.relationship(
                SubCategory,
                backref='categories',
                secondary=category_subcategory
            )
        return Category

    @pytest.fixture
    def SubCategory(self, Base, Product):

        subcategory_product = sa.Table(
            'subcategory_product',
            Base.metadata,
            sa.Column(
                'subcategory_id',
                sa.Integer,
                sa.ForeignKey('sub_category._id')
            ),
            sa.Column(
                'product_id',
                sa.Integer,
                sa.ForeignKey('product._id')
            )
        )

        class SubCategory(Base):
            __tablename__ = 'sub_category'
            id = sa.Column('_id', sa.Integer, primary_key=True)
            products = sa.orm.relationship(
                Product,
                backref='sub_categories',
                secondary=subcategory_product
            )
        return SubCategory

    @pytest.fixture
    def Product(self, Base):
        class Product(Base):
            __tablename__ = 'product'
            id = sa.Column('_id', sa.Integer, primary_key=True)
            price = sa.Column(sa.Numeric)
        return Product

    @pytest.fixture
    def init_models(self, Catalog, Category, SubCategory, Product):
        pass



import pytest
import sqlalchemy as sa
from sqlalchemy.orm import backref

from sqlalchemy_utils import auto_delete_orphans, ImproperlyConfigured


@pytest.fixture
def tagging_tbl(Base):
    return sa.Table(
        'tagging',
        Base.metadata,
        sa.Column(
            'tag_id',
            sa.Integer,
            sa.ForeignKey('tag.id', ondelete='cascade'),
            primary_key=True
        ),
        sa.Column(
            'entry_id',
            sa.Integer,
            sa.ForeignKey('entry.id', ondelete='cascade'),
            primary_key=True
        )
    )


@pytest.fixture
def Tag(Base):
    class Tag(Base):
        __tablename__ = 'tag'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.String(100), unique=True, nullable=False)

        def __init__(self, name=None):
            self.name = name
    return Tag


@pytest.fixture(
    params=['entries', backref('entries', lazy='select')],
    ids=['backref_string', 'backref_with_keywords']
)
def Entry(Base, Tag, tagging_tbl, request):
    class Entry(Base):
        __tablename__ = 'entry'

        id = sa.Column(sa.Integer, primary_key=True)

        tags = sa.orm.relationship(
            Tag,
            secondary=tagging_tbl,
            backref=request.param
        )
    auto_delete_orphans(Entry.tags)
    return Entry


@pytest.fixture
def EntryWithoutTagsBackref(Base, Tag, tagging_tbl):
    class EntryWithoutTagsBackref(Base):
        __tablename__ = 'entry'

        id = sa.Column(sa.Integer, primary_key=True)

        tags = sa.orm.relationship(
            Tag,
            secondary=tagging_tbl
        )
    return EntryWithoutTagsBackref


class TestAutoDeleteOrphans(object):

    @pytest.fixture
    def init_models(self, Entry, Tag):
        pass

    def test_orphan_deletion(self, session, Entry, Tag):
        r1 = Entry()
        r2 = Entry()
        r3 = Entry()
        t1, t2, t3, t4 = (
            Tag('t1'),
            Tag('t2'),
            Tag('t3'),
            Tag('t4')
        )

        r1.tags.extend([t1, t2])
        r2.tags.extend([t2, t3])
        r3.tags.extend([t4])
        session.add_all([r1, r2, r3])

        assert session.query(Tag).count() == 4
        r2.tags.remove(t2)
        assert session.query(Tag).count() == 4
        r1.tags.remove(t2)
        assert session.query(Tag).count() == 3
        r1.tags.remove(t1)
        assert session.query(Tag).count() == 2


class TestAutoDeleteOrphansWithoutBackref(object):

    @pytest.fixture
    def init_models(self, EntryWithoutTagsBackref, Tag):
        pass

    def test_orphan_deletion(self, EntryWithoutTagsBackref):
        with pytest.raises(ImproperlyConfigured):
            auto_delete_orphans(EntryWithoutTagsBackref.tags)



from decimal import Decimal

import pytest
import sqlalchemy as sa

from sqlalchemy_utils import NumericRangeType

intervals = None
inf = 0
try:
    import intervals
    from infinity import inf
except ImportError:
    pass


@pytest.fixture
def create_car(session, Car):
    def create_car(number_range):
        car = Car(
            price_range=number_range
        )

        session.add(car)
        session.commit()
        return session.query(Car).first()
    return create_car


@pytest.mark.skipif('intervals is None')
class NumericRangeTestCase(object):

    @pytest.fixture
    def Car(self, Base):
        class Car(Base):
            __tablename__ = 'car'
            id = sa.Column(sa.Integer, primary_key=True)
            price_range = sa.Column(NumericRangeType)

        return Car

    @pytest.fixture
    def init_models(self, Car):
        pass

    def test_nullify_range(self, create_car):
        car = create_car(None)
        assert car.price_range is None

    @pytest.mark.parametrize(
        'number_range',
        (
            [1, 3],
            (1, 3)
        )
    )
    def test_save_number_range(self, create_car, number_range):
        car = create_car(number_range)
        assert car.price_range.lower == 1
        assert car.price_range.upper == 3

    def test_infinite_upper_bound(self, create_car):
        car = create_car([1, inf])
        assert car.price_range.lower == 1
        assert car.price_range.upper == inf

    def test_infinite_lower_bound(self, create_car):
        car = create_car([-inf, 1])
        assert car.price_range.lower == -inf
        assert car.price_range.upper == 1

    def test_nullify_number_range(self, session, Car):
        car = Car(
            price_range=intervals.DecimalInterval([1, 3])
        )

        session.add(car)
        session.commit()

        car = session.query(Car).first()
        car.price_range = None
        session.commit()

        car = session.query(Car).first()
        assert car.price_range is None

    def test_integer_coercion(self, Car):
        car = Car(price_range=15)
        assert car.price_range.lower == 15
        assert car.price_range.upper == 15


@pytest.mark.usefixtures('postgresql_dsn')
class TestNumericRangeOnPostgres(NumericRangeTestCase):

    @pytest.mark.parametrize(
        ('number_range', 'length'),
        (
            ([1, 3], 2),
            ([1, 1], 0),
            ([-1, 1], 2),
            ([-inf, 1], None),
            ([0, inf], None),
            ([0, 0], 0),
            ([-3, -1], 2)
        )
    )
    def test_length(self, session, Car, create_car, number_range, length):
        create_car(number_range)
        query = (
            session.query(Car.price_range.length)
        )
        assert query.scalar() == length

    def test_literal_param(self, session, Car):
        clause = Car.price_range == [1, 3]
        compiled = str(clause.compile(compile_kwargs={'literal_binds': True}))
        assert compiled == "car.price_range = '[1, 3]'"


@pytest.mark.skipif('intervals is None')
class TestNumericRangeWithStep(object):
    @pytest.fixture
    def Car(self, Base):
        class Car(Base):
            __tablename__ = 'car'
            id = sa.Column(sa.Integer, primary_key=True)
            price_range = sa.Column(NumericRangeType(step=Decimal('0.5')))
        return Car

    @pytest.fixture
    def init_models(self, Car):
        pass

    def test_passes_step_argument_to_interval_object(self, create_car):
        car = create_car([Decimal('0.2'), Decimal('0.8')])
        assert car.price_range.lower == Decimal('0')
        assert car.price_range.upper == Decimal('1')
        assert car.price_range.step == Decimal('0.5')

    def test_passes_step_fetched_objects(self, session, Car, create_car):
        create_car([Decimal('0.2'), Decimal('0.8')])
        session.expunge_all()
        car = session.query(Car).first()
        assert car.price_range.lower == Decimal('0')
        assert car.price_range.upper == Decimal('1')
        assert car.price_range.step == Decimal('0.5')



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import has_changes


@pytest.fixture
def Article(Base):
    class Article(Base):
        __tablename__ = 'article_translation'
        id = sa.Column(sa.Integer, primary_key=True)
        title = sa.Column(sa.String(100))
    return Article


class TestHasChangesWithStringAttr(object):
    def test_without_changed_attr(self, Article):
        article = Article()
        assert not has_changes(article, 'title')

    def test_with_changed_attr(self, Article):
        article = Article(title='Some title')
        assert has_changes(article, 'title')


class TestHasChangesWithMultipleAttrs(object):
    def test_without_changed_attr(self, Article):
        article = Article()
        assert not has_changes(article, ['title'])

    def test_with_changed_attr(self, Article):
        article = Article(title='Some title')
        assert has_changes(article, ['title', 'id'])


class TestHasChangesWithExclude(object):
    def test_without_changed_attr(self, Article):
        article = Article()
        assert not has_changes(article, exclude=['id'])

    def test_with_changed_attr(self, Article):
        article = Article(title='Some title')
        assert has_changes(article, exclude=['id'])
        assert not has_changes(article, exclude=['title'])



import pytest
import sqlalchemy as sa

from sqlalchemy_utils.functions import (
    mock_engine,
    render_expression,
    render_statement
)


class TestRender(object):

    @pytest.fixture
    def User(self, Base):
        class User(Base):
            __tablename__ = 'user'
            id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
            name = sa.Column(sa.Unicode(255))
        return User

    @pytest.fixture
    def init_models(self, User):
        pass

    def test_render_orm_query(self, session, User):
        query = session.query(User).filter_by(id=3)
        text = render_statement(query)

        assert 'SELECT user.id, user.name' in text
        assert 'FROM user' in text
        assert 'WHERE user.id = 3' in text

    def test_render_statement(self, session, User):
        statement = User.__table__.select().where(User.id == 3)
        text = render_statement(statement, bind=session.bind)

        assert 'SELECT user.id, user.name' in text
        assert 'FROM user' in text
        assert 'WHERE user.id = 3' in text

    def test_render_statement_without_mapper(self, session):
        statement = sa.select([sa.text('1')])
        text = render_statement(statement, bind=session.bind)

        assert 'SELECT 1' in text

    def test_render_ddl(self, engine, User):
        expression = 'User.__table__.create(engine)'
        stream = render_expression(expression, engine)

        text = stream.getvalue()

        assert 'CREATE TABLE user' in text
        assert 'PRIMARY KEY' in text

    def test_render_mock_ddl(self, engine, User):
        # TODO: mock_engine doesn't seem to work with locally scoped variables.
        self.engine = engine
        with mock_engine('self.engine') as stream:
            User.__table__.create(self.engine)

        text = stream.getvalue()

        assert 'CREATE TABLE user' in text
        assert 'PRIMARY KEY' in text



import mock
import pytest
import sqlalchemy as sa
import sqlalchemy.dialects.mysql
import sqlalchemy.dialects.oracle
import sqlalchemy.dialects.postgresql
import sqlalchemy.dialects.sqlite
from sqlalchemy import inspect

from sqlalchemy_utils import Password, PasswordType, types  # noqa


@pytest.fixture
def extra_kwargs():
    """PasswordType extra keyword arguments."""
    return {}


@pytest.fixture
def User(Base, extra_kwargs):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, primary_key=True)
        password = sa.Column(PasswordType(
            schemes=[
                'pbkdf2_sha512',
                'pbkdf2_sha256',
                'md5_crypt',
                'hex_md5'
            ],
            deprecated=['md5_crypt', 'hex_md5'],
            **extra_kwargs
        ))

        def __repr__(self):
            return 'User(%r)' % self.id
    return User


@pytest.fixture
def init_models(User):
    pass


def onload_callback(schemes, deprecated):
    """
    Get onload callback that takes the PasswordType arguments from the config.
    """
    def onload(**kwargs):
        kwargs['schemes'] = schemes
        kwargs['deprecated'] = deprecated
        return kwargs
    return onload


@pytest.mark.skipif('types.password.passlib is None')
class TestPasswordType(object):

    @pytest.mark.parametrize('dialect_module,impl', [
        (sqlalchemy.dialects.sqlite, sa.dialects.sqlite.BLOB),
        (sqlalchemy.dialects.postgresql, sa.dialects.postgresql.BYTEA),
        (sqlalchemy.dialects.oracle, sa.dialects.oracle.RAW),
        (sqlalchemy.dialects.mysql, sa.VARBINARY),
    ])
    def test_load_dialect_impl(self, dialect_module, impl):
        """
        Should produce the same impl type as Alembic would expect after
        inspecing a database
        """
        password_type = PasswordType()
        assert isinstance(
            password_type.load_dialect_impl(dialect_module.dialect()),
            impl
        )

    def test_encrypt(self, User):
        """Should encrypt the password on setting the attribute."""
        obj = User()
        obj.password = b'b'

        assert obj.password.hash != 'b'
        assert obj.password.hash.startswith(b'$pbkdf2-sha512$')

    def test_check(self, session, User):
        """
        Should be able to compare the plaintext against the
        encrypted form.
        """
        obj = User()
        obj.password = 'b'

        assert obj.password == 'b'
        assert obj.password != 'a'

        session.add(obj)
        session.commit()

        obj = session.query(User).get(obj.id)

        assert obj.password == b'b'
        assert obj.password != 'a'

    def test_check_and_update(self, User):
        """
        Should be able to compare the plaintext against a deprecated
        encrypted form and have it auto-update to the preferred version.
        """

        from passlib.hash import md5_crypt

        obj = User()
        obj.password = Password(md5_crypt.hash('b'))

        assert obj.password.hash.decode('utf8').startswith('$1$')
        assert obj.password == 'b'
        assert obj.password.hash.decode('utf8').startswith('$pbkdf2-sha512$')

    def test_auto_column_length(self, User):
        """Should derive the correct column length from the specified schemes.
        """

        from passlib.hash import pbkdf2_sha512

        kind = inspect(User).c.password.type

        # name + rounds + salt + hash + ($ * 4) of largest hash
        expected_length = len(pbkdf2_sha512.name)
        expected_length += len(str(pbkdf2_sha512.max_rounds))
        expected_length += pbkdf2_sha512.max_salt_size
        expected_length += pbkdf2_sha512.encoded_checksum_size
        expected_length += 4

        assert kind.length == expected_length

    def test_without_schemes(self):
        assert PasswordType(schemes=[]).length == 1024

    def test_compare(self, User):
        from passlib.hash import md5_crypt

        obj = User()
        obj.password = Password(md5_crypt.hash('b'))

        other = User()
        other.password = Password(md5_crypt.hash('b'))

        # Not sure what to assert here; the test raised an error before.
        assert obj.password != other.password

    def test_set_none(self, session, User):

        obj = User()
        obj.password = None

        assert obj.password is None

        session.add(obj)
        session.commit()

        obj = session.query(User).get(obj.id)

        assert obj.password is None

    def test_update_none(self, session, User):
        """
        Should be able to change a password from ``None`` to a valid
        password.
        """

        obj = User()
        obj.password = None

        session.add(obj)
        session.commit()

        obj = session.query(User).get(obj.id)
        obj.password = 'b'

        session.commit()

    def test_compare_none(self, User):
        """
        Should be able to compare a password of ``None``.
        """

        obj = User()
        obj.password = None

        assert obj.password is None
        assert obj.password == None  # noqa

        obj.password = 'b'

        assert obj.password is not None
        assert obj.password != None  # noqa

    def test_check_and_update_persist(self, session, User):
        """
        When a password is compared, the hash should update if needed to
        change the algorithm; and, commit to the database.
        """

        from passlib.hash import md5_crypt

        obj = User()
        obj.password = Password(md5_crypt.hash('b'))

        session.add(obj)
        session.commit()

        assert obj.password.hash.decode('utf8').startswith('$1$')
        assert obj.password == 'b'

        session.commit()

        obj = session.query(User).get(obj.id)

        assert obj.password.hash.decode('utf8').startswith('$pbkdf2-sha512$')
        assert obj.password == 'b'

    @pytest.mark.parametrize(
        'extra_kwargs',
        [
            dict(
                onload=onload_callback(
                    schemes=['pbkdf2_sha256'],
                    deprecated=[],
                )
            )
        ]
    )
    def test_lazy_configuration(self, User):
        """
        Field should be able to read the passlib attributes lazily from the
        config (e.g. Flask config).
        """
        schemes = User.password.type.context.schemes()
        assert tuple(schemes) == ('pbkdf2_sha256',)
        obj = User()
        obj.password = b'b'
        assert obj.password.hash.decode('utf8').startswith('$pbkdf2-sha256$')

    @pytest.mark.parametrize('max_length', [1, 103])
    def test_constant_length(self, max_length):
        """
        Test that constant max_length is applied.
        """
        typ = PasswordType(max_length=max_length)
        assert typ.length == max_length

    def test_context_is_lazy(self):
        """
        Make sure the init doesn't evaluate the lazy context.
        """
        onload = mock.Mock(return_value={})
        PasswordType(onload=onload)
        assert not onload.called



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import get_type


@pytest.fixture
def User(Base):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, primary_key=True)
    return User


@pytest.fixture
def Article(Base, User):
    class Article(Base):
        __tablename__ = 'article'
        id = sa.Column(sa.Integer, primary_key=True)

        author_id = sa.Column(sa.Integer, sa.ForeignKey(User.id))
        author = sa.orm.relationship(User)

        some_property = sa.orm.column_property(
            sa.func.coalesce(id, 1)
        )
    return Article


class TestGetType(object):

    def test_instrumented_attribute(self, Article):
        assert isinstance(get_type(Article.id), sa.Integer)

    def test_column_property(self, Article):
        assert isinstance(get_type(Article.id.property), sa.Integer)

    def test_column(self, Article):
        assert isinstance(get_type(Article.__table__.c.id), sa.Integer)

    def test_calculated_column_property(self, Article):
        assert isinstance(get_type(Article.some_property), sa.Integer)

    def test_relationship_property(self, Article, User):
        assert get_type(Article.author) == User

    def test_scalar_select(self, Article):
        query = sa.select([Article.id]).as_scalar()
        assert isinstance(get_type(query), sa.Integer)



# -*- coding: utf-8 -*-
import pytest
import sqlalchemy as sa

from sqlalchemy_utils import i18n
from sqlalchemy_utils.primitives import WeekDays
from sqlalchemy_utils.types import WeekDaysType


@pytest.fixture
def Schedule(Base):
    class Schedule(Base):
        __tablename__ = 'schedule'
        id = sa.Column(sa.Integer, primary_key=True)
        working_days = sa.Column(WeekDaysType)

        def __repr__(self):
            return 'Schedule(%r)' % self.id
    return Schedule


@pytest.fixture
def init_models(Schedule):
    pass


@pytest.fixture
def set_get_locale():
    i18n.get_locale = lambda: i18n.babel.Locale('en')


@pytest.mark.usefixtures('set_get_locale')
@pytest.mark.skipif('i18n.babel is None')
class WeekDaysTypeTestCase(object):

    def test_color_parameter_processing(self, session, Schedule):
        schedule = Schedule(
            working_days=b'0001111'
        )
        session.add(schedule)
        session.commit()

        schedule = session.query(Schedule).first()
        assert isinstance(schedule.working_days, WeekDays)

    def test_scalar_attributes_get_coerced_to_objects(self, Schedule):
        schedule = Schedule(working_days=b'1010101')

        assert isinstance(schedule.working_days, WeekDays)


@pytest.mark.usefixtures('sqlite_memory_dsn')
class TestWeekDaysTypeOnSQLite(WeekDaysTypeTestCase):
    pass


@pytest.mark.usefixtures('postgresql_dsn')
class TestWeekDaysTypeOnPostgres(WeekDaysTypeTestCase):
    pass


@pytest.mark.usefixtures('mysql_dsn')
class TestWeekDaysTypeOnMySQL(WeekDaysTypeTestCase):
    pass



import pytest

from sqlalchemy_utils.types.encrypted.padding import (
    InvalidPaddingError,
    PKCS5Padding
)


class TestPkcs5Padding(object):
    def setup_method(self):
        self.BLOCK_SIZE = 8
        self.padder = PKCS5Padding(self.BLOCK_SIZE)

    def test_various_lengths_roundtrip(self):
        for l in range(0, 3 * self.BLOCK_SIZE):
            val = b'*' * l
            padded = self.padder.pad(val)
            unpadded = self.padder.unpad(padded)
            assert val == unpadded, 'Round trip error for length %d' % l

    def test_invalid_unpad(self):
        with pytest.raises(InvalidPaddingError):
            self.padder.unpad(None)
        with pytest.raises(InvalidPaddingError):
            self.padder.unpad(b'')
        with pytest.raises(InvalidPaddingError):
            self.padder.unpad(b'\01')
        with pytest.raises(InvalidPaddingError):
            self.padder.unpad((b'*' * (self.BLOCK_SIZE - 1)) + b'\00')
        with pytest.raises(InvalidPaddingError):
            self.padder.unpad((b'*' * self.BLOCK_SIZE) + b'\01')

    def test_pad_longer_than_block(self):
        with pytest.raises(InvalidPaddingError):
            self.padder.unpad(
                'x' * (self.BLOCK_SIZE - 1) +
                chr(self.BLOCK_SIZE + 1) * (self.BLOCK_SIZE + 1)
            )

    def test_incorrect_padding(self):
        # Hard-coded for blocksize of 8
        assert self.padder.unpad(b'1234\04\04\04\04') == b'1234'
        with pytest.raises(InvalidPaddingError):
            self.padder.unpad(b'1234\02\04\04\04')
        with pytest.raises(InvalidPaddingError):
            self.padder.unpad(b'1234\04\02\04\04')
        with pytest.raises(InvalidPaddingError):
            self.padder.unpad(b'1234\04\04\02\04')



import os
import warnings

import pytest
import sqlalchemy as sa
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base, synonym_for
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import close_all_sessions
from sqlalchemy_utils import (
    aggregates,
    coercion_listener,
    i18n,
    InstrumentedList
)

from sqlalchemy_utils.types.pg_composite import remove_composite_listeners


@sa.event.listens_for(sa.engine.Engine, 'before_cursor_execute')
def count_sql_calls(conn, cursor, statement, parameters, context, executemany):
    try:
        conn.query_count += 1
    except AttributeError:
        conn.query_count = 0


warnings.simplefilter('error', sa.exc.SAWarning)

sa.event.listen(sa.orm.mapper, 'mapper_configured', coercion_listener)


def get_locale():
    class Locale():
        territories = {'FI': 'Finland'}

    return Locale()


@pytest.fixture(scope='session')
def db_name():
    return os.environ.get('SQLALCHEMY_UTILS_TEST_DB', 'sqlalchemy_utils_test')


@pytest.fixture(scope='session')
def postgresql_db_user():
    return os.environ.get('SQLALCHEMY_UTILS_TEST_POSTGRESQL_USER', 'postgres')


@pytest.fixture(scope='session')
def mysql_db_user():
    return os.environ.get('SQLALCHEMY_UTILS_TEST_MYSQL_USER', 'root')


@pytest.fixture
def postgresql_dsn(postgresql_db_user, db_name):
    return 'postgresql://{0}@localhost/{1}'.format(postgresql_db_user, db_name)


@pytest.fixture
def mysql_dsn(mysql_db_user, db_name):
    return 'mysql+pymysql://{0}@localhost/{1}'.format(mysql_db_user, db_name)


@pytest.fixture
def sqlite_memory_dsn():
    return 'sqlite:///:memory:'


@pytest.fixture
def sqlite_none_database_dsn():
    return 'sqlite://'


@pytest.fixture
def sqlite_file_dsn(db_name):
    return 'sqlite:///{0}.db'.format(db_name)


@pytest.fixture
def mssql_db_user():
    return os.environ.get('SQLALCHEMY_UTILS_TEST_MSSQL_USER', 'sa')


@pytest.fixture
def mssql_db_password():
    return os.environ.get('SQLALCHEMY_UTILS_TEST_MSSQL_PASSWORD',
                          'Strong!Passw0rd')


@pytest.fixture
def mssql_db_driver():
    driver = os.environ.get('SQLALCHEMY_UTILS_TEST_MSSQL_DRIVER',
                            'ODBC Driver 17 for SQL Server')
    return driver.replace(' ', '+')


@pytest.fixture
def mssql_dsn(mssql_db_user, mssql_db_password, mssql_db_driver, db_name):
    return 'mssql+pyodbc://{0}:{1}@localhost/{2}?driver={3}'\
        .format(mssql_db_user, mssql_db_password, db_name, mssql_db_driver)


@pytest.fixture
def dsn(request):
    if 'postgresql_dsn' in request.fixturenames:
        return request.getfixturevalue('postgresql_dsn')
    elif 'mysql_dsn' in request.fixturenames:
        return request.getfixturevalue('mysql_dsn')
    elif 'mssql_dsn' in request.fixturenames:
        return request.getfixturevalue('mssql_dsn')
    elif 'sqlite_file_dsn' in request.fixturenames:
        return request.getfixturevalue('sqlite_file_dsn')
    elif 'sqlite_memory_dsn' in request.fixturenames:
        pass  # Return default
    return request.getfixturevalue('sqlite_memory_dsn')


@pytest.fixture
def engine(dsn):
    engine = create_engine(dsn)
    # engine.echo = True
    return engine


@pytest.fixture
def connection(engine):
    return engine.connect()


@pytest.fixture
def Base():
    return declarative_base()


@pytest.fixture
def User(Base):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
        name = sa.Column(sa.Unicode(255))
    return User


@pytest.fixture
def Category(Base):

    class Category(Base):
        __tablename__ = 'category'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))
        title = sa.Column(sa.Unicode(255))

        @hybrid_property
        def full_name(self):
            return u'%s %s' % (self.title, self.name)

        @full_name.expression
        def full_name(self):
            return sa.func.concat(self.title, ' ', self.name)

        @hybrid_property
        def articles_count(self):
            return len(self.articles)

        @articles_count.expression
        def articles_count(cls):
            Article = Base._decl_class_registry['Article']
            return (
                sa.select([sa.func.count(Article.id)])
                .where(Article.category_id == cls.id)
                .correlate(Article.__table__)
                .label('article_count')
            )

        @property
        def name_alias(self):
            return self.name

        @synonym_for('name')
        @property
        def name_synonym(self):
            return self.name
    return Category


@pytest.fixture
def Article(Base, Category):
    class Article(Base):
        __tablename__ = 'article'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255), index=True)
        category_id = sa.Column(sa.Integer, sa.ForeignKey(Category.id))

        category = sa.orm.relationship(
            Category,
            primaryjoin=category_id == Category.id,
            backref=sa.orm.backref(
                'articles',
                collection_class=InstrumentedList
            )
        )
    return Article


@pytest.fixture
def init_models(User, Category, Article):
    pass


@pytest.fixture
def session(request, engine, connection, Base, init_models):
    sa.orm.configure_mappers()
    Base.metadata.create_all(connection)
    Session = sessionmaker(bind=connection)
    session = Session()
    i18n.get_locale = get_locale

    def teardown():
        aggregates.manager.reset()
        close_all_sessions()
        Base.metadata.drop_all(connection)
        remove_composite_listeners()
        connection.close()
        engine.dispose()

    request.addfinalizer(teardown)

    return session



from datetime import datetime, timedelta

import pytest
import sqlalchemy as sa

from sqlalchemy_utils import DateTimeRangeType

intervals = None
inf = 0
try:
    import intervals
    from infinity import inf
except ImportError:
    pass


@pytest.fixture
def Booking(Base):
    class Booking(Base):
        __tablename__ = 'booking'
        id = sa.Column(sa.Integer, primary_key=True)
        during = sa.Column(DateTimeRangeType)

    return Booking


@pytest.fixture
def create_booking(session, Booking):
    def create_booking(date_range):
        booking = Booking(
            during=date_range
        )
        session.add(booking)
        session.commit()
        return session.query(Booking).first()
    return create_booking


@pytest.fixture
def init_models(Booking):
    pass


@pytest.mark.skipif('intervals is None')
class DateRangeTestCase(object):

    def test_nullify_range(self, create_booking):
        booking = create_booking(None)
        assert booking.during is None

    @pytest.mark.parametrize(
        ('date_range'),
        (
            [datetime(2015, 1, 1), datetime(2015, 1, 3)],
            [datetime(2015, 1, 1), inf],
            [-inf, datetime(2015, 1, 1)]
        )
    )
    def test_save_date_range(self, create_booking, date_range):
        booking = create_booking(date_range)
        assert booking.during.lower == date_range[0]
        assert booking.during.upper == date_range[1]

    def test_nullify_date_range(self, session, Booking):
        booking = Booking(
            during=intervals.DateInterval(
                [datetime(2015, 1, 1), datetime(2015, 1, 3)]
            )
        )

        session.add(booking)
        session.commit()

        booking = session.query(Booking).first()
        booking.during = None
        session.commit()

        booking = session.query(Booking).first()
        assert booking.during is None

    def test_integer_coercion(self, Booking):
        booking = Booking(during=datetime(2015, 1, 1))
        assert booking.during.lower == datetime(2015, 1, 1)
        assert booking.during.upper == datetime(2015, 1, 1)


@pytest.mark.usefixtures('postgresql_dsn')
class TestDateRangeOnPostgres(object):

    @pytest.mark.parametrize(
        ('date_range', 'length'),
        (
            (
                [datetime(2015, 1, 1), datetime(2015, 1, 3)],
                timedelta(days=2)
            ),
            (
                [datetime(2015, 1, 1), datetime(2015, 1, 1)],
                timedelta(days=0)
            ),
            ([-inf, datetime(2015, 1, 1)], None),
            ([datetime(2015, 1, 1), inf], None),
        )
    )
    def test_length(
        self,
        session,
        Booking,
        create_booking,
        date_range,
        length
    ):
        create_booking(date_range)
        query = (
            session.query(Booking.during.length)
        )
        assert query.scalar() == length



import pytest
import sqlalchemy as sa

from sqlalchemy_utils.types import locale


@pytest.fixture
def User(Base):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, primary_key=True)
        locale = sa.Column(locale.LocaleType)

        def __repr__(self):
            return 'User(%r)' % self.id
    return User


@pytest.fixture
def init_models(User):
    pass


@pytest.mark.skipif('locale.babel is None')
class TestLocaleType(object):

    def test_parameter_processing(self, session, User):
        user = User(
            locale=locale.babel.Locale(u'fi')
        )

        session.add(user)
        session.commit()

        user = session.query(User).first()

    def test_territory_parsing(self, session, User):
        ko_kr = locale.babel.Locale(u'ko', territory=u'KR')
        user = User(locale=ko_kr)
        session.add(user)
        session.commit()

        assert session.query(User.locale).first()[0] == ko_kr

    def test_coerce_territory_parsing(self, User):
        user = User()
        user.locale = 'ko_KR'
        assert user.locale == locale.babel.Locale(u'ko', territory=u'KR')

    def test_scalar_attributes_get_coerced_to_objects(self, User):
        user = User(locale='en_US')

        assert isinstance(user.locale, locale.babel.Locale)

    def test_unknown_locale_throws_exception(self, User):
        with pytest.raises(locale.babel.UnknownLocaleError):
            User(locale=u'unknown')

    def test_literal_param(self, session, User):
        clause = User.locale == 'en_US'
        compiled = str(clause.compile(compile_kwargs={'literal_binds': True}))
        assert compiled == '"user".locale = \'en_US\''



import pytest
import six
import sqlalchemy as sa

from sqlalchemy_utils import (  # noqa
    PhoneNumber,
    PhoneNumberParseException,
    PhoneNumberType,
    types
)

VALID_PHONE_NUMBERS = (
    '040 1234567',
    '+358 401234567',
    '09 2501234',
    '+358 92501234',
    '0800 939393',
    '09 4243 0456',
    '0600 900 500'
)


@pytest.fixture
def User(Base):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
        name = sa.Column(sa.Unicode(255))
        phone_number = sa.Column(PhoneNumberType())
    return User


@pytest.fixture
def init_models(User):
    pass


@pytest.fixture
def phone_number():
    return PhoneNumber(
        '040 1234567',
        'FI'
    )


@pytest.fixture
def user(session, User, phone_number):
    user = User()
    user.name = u'Someone'
    user.phone_number = phone_number
    session.add(user)
    session.commit()
    return user


@pytest.mark.skipif('types.phone_number.phonenumbers is None')
class TestPhoneNumber(object):

    @pytest.mark.parametrize('raw_number', VALID_PHONE_NUMBERS)
    def test_valid_phone_numbers(self, raw_number):
        number = PhoneNumber(raw_number, 'FI')
        assert number.is_valid_number()

    @pytest.mark.parametrize('raw_number', ('abc', '+040 1234567'))
    def test_invalid_phone_numbers__constructor_fails(self, raw_number):
        with pytest.raises(PhoneNumberParseException):
            PhoneNumber(raw_number, 'FI')

    @pytest.mark.parametrize('raw_number', ('0111234567', '358'))
    def test_invalid_phone_numbers__is_valid_number(self, raw_number):
        number = PhoneNumber(raw_number, 'FI')
        assert not number.is_valid_number()

    def test_invalid_phone_numbers_throw_dont_wrap_exception(
        self,
        session,
        User
    ):
        with pytest.raises(PhoneNumberParseException):
            session.execute(
                User.__table__.insert().values(
                    name=u'Someone',
                    phone_number=u'abc'
                )
            )

    def test_phone_number_attributes(self):
        number = PhoneNumber('+358401234567')
        assert number.e164 == u'+358401234567'
        assert number.international == u'+358 40 1234567'
        assert number.national == u'040 1234567'

    def test_phone_number_attributes_for_short_code(self):
        """
        For international and national shortcode remains the same, if we pass
        short code to PhoneNumber library without giving check_region it will
        raise exception
        :return:
        """
        number = PhoneNumber('72404', check_region=False)
        assert number.e164 == u'+072404'
        assert number.international == u'72404'
        assert number.national == u'72404'

    def test_phone_number_str_repr(self):
        number = PhoneNumber('+358401234567')
        if six.PY2:
            assert unicode(number) == number.national  # noqa
            assert str(number) == number.national.encode('utf-8')
        else:
            assert str(number) == number.national


@pytest.mark.skipif('types.phone_number.phonenumbers is None')
class TestPhoneNumberType(object):

    def test_query_returns_phone_number_object(
        self,
        session,
        User,
        user,
        phone_number
    ):
        queried_user = session.query(User).first()
        assert queried_user.phone_number == phone_number

    def test_phone_number_is_stored_as_string(self, session, user):
        result = session.execute(
            'SELECT phone_number FROM "user" WHERE id=:param',
            {'param': user.id}
        )
        assert result.first()[0] == u'+358401234567'

    def test_phone_number_with_extension(self, session, User):
        user = User(phone_number='555-555-5555 Ext. 555')

        session.add(user)
        session.commit()
        session.refresh(user)
        assert user.phone_number.extension == '555'

    def test_empty_phone_number_is_equiv_to_none(self, session, User):
        user = User(phone_number='')

        session.add(user)
        session.commit()
        session.refresh(user)
        assert user.phone_number is None

    def test_uses_phonenumber_class_as_python_type(self):
        assert PhoneNumberType().python_type is PhoneNumber

    @pytest.mark.usefixtures('user')
    def test_phone_number_is_none(self, session, User):
        phone_number = None
        user = User()
        user.name = u'Someone'
        user.phone_number = phone_number
        session.add(user)
        session.commit()
        queried_user = session.query(User)[1]
        assert queried_user.phone_number is None
        result = session.execute(
            'SELECT phone_number FROM "user" WHERE id=:param',
            {'param': user.id}
        )
        assert result.first()[0] is None

    def test_scalar_attributes_get_coerced_to_objects(self, User):
        user = User(phone_number='050111222')

        assert isinstance(user.phone_number, PhoneNumber)


@pytest.mark.skipif('types.phone_number.phonenumbers is None')
class TestPhoneNumberComposite(object):
    @pytest.fixture
    def User(self, Base):
        class User(Base):
            __tablename__ = 'user'
            id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
            name = sa.Column(sa.String(255))
            _phone_number = sa.Column(sa.String(255))
            country = sa.Column(sa.String(255))
            phone_number = sa.orm.composite(
                PhoneNumber,
                _phone_number,
                country
            )
        return User

    @pytest.fixture
    def user(self, session, User):
        user = User()
        user.name = u'Someone'
        user.phone_number = PhoneNumber('+35840111222', 'FI')
        session.add(user)
        session.commit()
        return user

    def test_query_returns_phone_number_object(
        self,
        session,
        User,
        user
    ):
        queried_user = session.query(User).first()
        assert queried_user.phone_number.national == '040 111222'
        assert queried_user.phone_number.region == 'FI'



from __future__ import absolute_import

from datetime import datetime

import six
from sqlalchemy import types

from ..exceptions import ImproperlyConfigured
from .scalar_coercible import ScalarCoercible

try:
    from collections.abc import Iterable
except ImportError:  # For python 2.7 support
    from collections import Iterable

arrow = None
try:
    import arrow
except ImportError:
    pass


class ArrowType(types.TypeDecorator, ScalarCoercible):
    """
    ArrowType provides way of saving Arrow_ objects into database. It
    automatically changes Arrow_ objects to datetime objects on the way in and
    datetime objects back to Arrow_ objects on the way out (when querying
    database). ArrowType needs Arrow_ library installed.

    .. _Arrow: http://crsmithdev.com/arrow/

    ::

        from datetime import datetime
        from sqlalchemy_utils import ArrowType
        import arrow


        class Article(Base):
            __tablename__ = 'article'
            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.Unicode(255))
            created_at = sa.Column(ArrowType)



        article = Article(created_at=arrow.utcnow())


    As you may expect all the arrow goodies come available:

    ::


        article.created_at = article.created_at.replace(hours=-1)

        article.created_at.humanize()
        # 'an hour ago'

    """
    impl = types.DateTime

    def __init__(self, *args, **kwargs):
        if not arrow:
            raise ImproperlyConfigured(
                "'arrow' package is required to use 'ArrowType'"
            )

        super(ArrowType, self).__init__(*args, **kwargs)

    def process_bind_param(self, value, dialect):
        if value:
            utc_val = self._coerce(value).to('UTC')
            return utc_val.datetime if self.impl.timezone else utc_val.naive
        return value

    def process_result_value(self, value, dialect):
        if value:
            return arrow.get(value)
        return value

    def process_literal_param(self, value, dialect):
        return str(value)

    def _coerce(self, value):
        if value is None:
            return None
        elif isinstance(value, six.string_types):
            value = arrow.get(value)
        elif isinstance(value, Iterable):
            value = arrow.get(*value)
        elif isinstance(value, datetime):
            value = arrow.get(value)
        return value

    @property
    def python_type(self):
        return self.impl.type.python_type



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import aggregated


@pytest.fixture
def Category(Base):
    class Category(Base):
        __tablename__ = 'category'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))
    return Category


@pytest.fixture
def Catalog(Base, Category):
    class Catalog(Base):
        __tablename__ = 'catalog'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))

        @aggregated(
            'products.categories',
            sa.Column(sa.Integer, default=0)
        )
        def category_count(self):
            return sa.func.count(sa.distinct(Category.id))
    return Catalog


@pytest.fixture
def Product(Base, Catalog, Category):
    product_categories = sa.Table(
        'category_product',
        Base.metadata,
        sa.Column('category_id', sa.Integer, sa.ForeignKey('category.id')),
        sa.Column('product_id', sa.Integer, sa.ForeignKey('product.id'))
    )

    class Product(Base):
        __tablename__ = 'product'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.Unicode(255))
        price = sa.Column(sa.Numeric)

        catalog_id = sa.Column(
            sa.Integer, sa.ForeignKey('catalog.id')
        )

        catalog = sa.orm.relationship(
            Catalog,
            backref='products'
        )

        categories = sa.orm.relationship(
            Category,
            backref='products',
            secondary=product_categories
        )
    return Product


@pytest.fixture
def init_models(Category, Catalog, Product):
    pass


@pytest.mark.usefixtures('postgresql_dsn')
class TestAggregateOneToManyAndManyToMany(object):

    def test_insert(self, session, Category, Catalog, Product):
        category = Category()
        products = [
            Product(categories=[category]),
            Product(categories=[category])
        ]
        catalog = Catalog(products=products)
        session.add(catalog)
        products2 = [
            Product(categories=[category]),
            Product(categories=[category])
        ]
        catalog2 = Catalog(products=products2)
        session.add(catalog)
        session.commit()
        assert catalog.category_count == 1
        assert catalog2.category_count == 1



import uuid

import pytest
import sqlalchemy as sa

from sqlalchemy_utils import UUIDType


@pytest.fixture
def User(Base):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(UUIDType, default=uuid.uuid4, primary_key=True)

        def __repr__(self):
            return 'User(%r)' % self.id
    return User


@pytest.fixture
def init_models(User):
    pass


class TestUUIDType(object):

    def test_repr(self):
        plain = UUIDType()
        assert repr(plain) == 'UUIDType()'

        text = UUIDType(binary=False)
        assert repr(text) == 'UUIDType(binary=False)'

        not_native = UUIDType(native=False)
        assert repr(not_native) == 'UUIDType(native=False)'

    def test_commit(self, session, User):
        obj = User()
        obj.id = uuid.uuid4().hex

        session.add(obj)
        session.commit()

        u = session.query(User).one()

        assert u.id == obj.id

    def test_coerce(self, User):
        obj = User()
        obj.id = identifier = uuid.uuid4().hex

        assert isinstance(obj.id, uuid.UUID)
        assert obj.id.hex == identifier

        obj.id = identifier = uuid.uuid4().bytes

        assert isinstance(obj.id, uuid.UUID)
        assert obj.id.bytes == identifier



import pytest
import six
import sqlalchemy as sa

from sqlalchemy_utils import generic_relationship

from ..generic_relationship import GenericRelationshipTestCase


@pytest.fixture
def incrementor():
    class Incrementor(object):
        value = 1
    return Incrementor()


@pytest.fixture
def Building(Base, incrementor):
    class Building(Base):
        __tablename__ = 'building'
        id = sa.Column(sa.Integer, primary_key=True)
        code = sa.Column(sa.Integer, primary_key=True)

        def __init__(self):
            incrementor.value += 1
            self.id = incrementor.value
            self.code = incrementor.value
    return Building


@pytest.fixture
def User(Base, incrementor):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, primary_key=True)
        code = sa.Column(sa.Integer, primary_key=True)

        def __init__(self):
            incrementor.value += 1
            self.id = incrementor.value
            self.code = incrementor.value
    return User


@pytest.fixture
def Event(Base):
    class Event(Base):
        __tablename__ = 'event'
        id = sa.Column(sa.Integer, primary_key=True)

        object_type = sa.Column(sa.Unicode(255))
        object_id = sa.Column(sa.Integer, nullable=False)
        object_code = sa.Column(sa.Integer, nullable=False)

        object = generic_relationship(
            object_type, (object_id, object_code)
        )
    return Event


@pytest.fixture
def init_models(Building, User, Event):
    pass


class TestGenericRelationship(GenericRelationshipTestCase):

    def test_set_manual_and_get(self, session, Event, User):
        user = User()

        session.add(user)
        session.commit()

        event = Event()
        event.object_id = user.id
        event.object_type = six.text_type(type(user).__name__)
        event.object_code = user.code

        assert event.object is None

        session.add(event)
        session.commit()

        assert event.object == user



# -*- coding: utf-8 -*-
import pytest
import sqlalchemy as sa
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import close_all_sessions

from sqlalchemy_utils import (
    CompositeArray,
    CompositeType,
    Currency,
    CurrencyType,
    i18n,
    NumericRangeType,
    register_composites,
    remove_composite_listeners
)
from sqlalchemy_utils.types import pg_composite
from sqlalchemy_utils.types.range import intervals


@pytest.mark.usefixtures('postgresql_dsn')
class TestCompositeTypeWithRegularTypes(object):

    @pytest.fixture
    def Account(self, Base):
        class Account(Base):
            __tablename__ = 'account'
            id = sa.Column(sa.Integer, primary_key=True)
            balance = sa.Column(
                CompositeType(
                    'money_type',
                    [
                        sa.Column('currency', sa.String),
                        sa.Column('amount', sa.Integer)
                    ]
                )
            )

        return Account

    @pytest.fixture
    def init_models(self, Account):
        pass

    def test_parameter_processing(self, session, Account):
        account = Account(
            balance=('USD', 15)
        )

        session.add(account)
        session.commit()

        account = session.query(Account).first()
        assert account.balance.currency == 'USD'
        assert account.balance.amount == 15

    def test_non_ascii_chars(self, session, Account):
        account = Account(
            balance=(u'ääöö', 15)
        )

        session.add(account)
        session.commit()

        account = session.query(Account).first()
        assert account.balance.currency == u'ääöö'
        assert account.balance.amount == 15


@pytest.mark.skipif('i18n.babel is None')
@pytest.mark.usefixtures('postgresql_dsn')
class TestCompositeTypeWithTypeDecorators(object):

    @pytest.fixture
    def Account(self, Base):
        class Account(Base):
            __tablename__ = 'account'
            id = sa.Column(sa.Integer, primary_key=True)
            balance = sa.Column(
                CompositeType(
                    'money_type',
                    [
                        sa.Column('currency', CurrencyType),
                        sa.Column('amount', sa.Integer)
                    ]
                )
            )

        return Account

    @pytest.fixture
    def init_models(self, Account):
        i18n.get_locale = lambda: i18n.babel.Locale('en')

    def test_result_set_processing(self, session, Account):
        account = Account(
            balance=('USD', 15)
        )

        session.add(account)
        session.commit()

        account = session.query(Account).first()
        assert account.balance.currency == Currency('USD')
        assert account.balance.amount == 15

    def test_parameter_processing(self, session, Account):
        account = Account(
            balance=(Currency('USD'), 15)
        )

        session.add(account)
        session.commit()

        account = session.query(Account).first()
        assert account.balance.currency == Currency('USD')
        assert account.balance.amount == 15


@pytest.mark.skipif('i18n.babel is None')
@pytest.mark.usefixtures('postgresql_dsn')
class TestCompositeTypeInsideArray(object):

    @pytest.fixture
    def type_(self):
        return CompositeType(
            'money_type',
            [
                sa.Column('currency', CurrencyType),
                sa.Column('amount', sa.Integer)
            ]
        )

    @pytest.fixture
    def Account(self, Base, type_):
        class Account(Base):
            __tablename__ = 'account'
            id = sa.Column(sa.Integer, primary_key=True)
            balances = sa.Column(
                CompositeArray(type_)
            )

        return Account

    @pytest.fixture
    def init_models(self, Account):
        i18n.get_locale = lambda: i18n.babel.Locale('en')

    def test_parameter_processing(self, session, type_, Account):
        account = Account(
            balances=[
                type_.type_cls(Currency('USD'), 15),
                type_.type_cls(Currency('AUD'), 20)
            ]
        )

        session.add(account)
        session.commit()

        account = session.query(Account).first()
        assert account.balances[0].currency == Currency('USD')
        assert account.balances[0].amount == 15
        assert account.balances[1].currency == Currency('AUD')
        assert account.balances[1].amount == 20


@pytest.mark.skipif('intervals is None')
@pytest.mark.usefixtures('postgresql_dsn')
class TestCompositeTypeWithRangeTypeInsideArray(object):

    @pytest.fixture
    def type_(self):
        return CompositeType(
            'category',
            [
                sa.Column('scale', NumericRangeType),
                sa.Column('name', sa.String)
            ]
        )

    @pytest.fixture
    def Account(self, Base, type_):
        class Account(Base):
            __tablename__ = 'account'
            id = sa.Column(sa.Integer, primary_key=True)
            categories = sa.Column(
                CompositeArray(type_)
            )

        return Account

    @pytest.fixture
    def init_models(self, Account):
        pass

    def test_parameter_processing_with_named_tuple(
        self,
        session,
        type_,
        Account
    ):
        account = Account(
            categories=[
                type_.type_cls(
                    intervals.DecimalInterval([15, 18]),
                    'bad'
                ),
                type_.type_cls(
                    intervals.DecimalInterval([18, 20]),
                    'good'
                )
            ]
        )

        session.add(account)
        session.commit()

        account = session.query(Account).first()
        assert (
            account.categories[0].scale == intervals.DecimalInterval([15, 18])
        )
        assert account.categories[0].name == 'bad'
        assert (
            account.categories[1].scale == intervals.DecimalInterval([18, 20])
        )
        assert account.categories[1].name == 'good'

    def test_parameter_processing_with_tuple(self, session, Account):
        account = Account(
            categories=[
                (intervals.DecimalInterval([15, 18]), 'bad'),
                (intervals.DecimalInterval([18, 20]), 'good')
            ]
        )

        session.add(account)
        session.commit()

        account = session.query(Account).first()
        assert (
            account.categories[0].scale == intervals.DecimalInterval([15, 18])
        )
        assert account.categories[0].name == 'bad'
        assert (
            account.categories[1].scale == intervals.DecimalInterval([18, 20])
        )
        assert account.categories[1].name == 'good'

    def test_parameter_processing_with_nulls_as_composite_fields(
        self,
        session,
        Account
    ):
        account = Account(
            categories=[
                (None, 'bad'),
                (intervals.DecimalInterval([18, 20]), None)
            ]
        )
        session.add(account)
        session.commit()
        assert account.categories[0].scale is None
        assert account.categories[0].name == 'bad'
        assert (
            account.categories[1].scale == intervals.DecimalInterval([18, 20])
        )
        assert account.categories[1].name is None

    def test_parameter_processing_with_nulls_as_composites(
        self,
        session,
        Account
    ):
        account = Account(
            categories=[
                (None, None),
                None
            ]
        )
        session.add(account)
        session.commit()
        assert account.categories[0].scale is None
        assert account.categories[0].name is None
        assert account.categories[1] is None


@pytest.mark.usefixtures('postgresql_dsn')
class TestCompositeTypeWhenTypeAlreadyExistsInDatabase(object):

    @pytest.fixture
    def Account(self, Base):
        pg_composite.registered_composites = {}

        type_ = CompositeType(
            'money_type',
            [
                sa.Column('currency', sa.String),
                sa.Column('amount', sa.Integer)
            ]
        )

        class Account(Base):
            __tablename__ = 'account'
            id = sa.Column(sa.Integer, primary_key=True)
            balance = sa.Column(type_)

        return Account

    @pytest.fixture
    def session(self, request, engine, connection, Base, Account):
        sa.orm.configure_mappers()

        Session = sessionmaker(bind=connection)
        session = Session()
        session.execute(
            "CREATE TYPE money_type AS (currency VARCHAR, amount INTEGER)"
        )
        session.execute(
            """CREATE TABLE account (
                id SERIAL, balance MONEY_TYPE, PRIMARY KEY(id)
            )"""
        )

        def teardown():
            session.execute('DROP TABLE account')
            session.execute('DROP TYPE money_type')
            session.commit()
            close_all_sessions()
            connection.close()
            remove_composite_listeners()
            engine.dispose()

        register_composites(connection)
        request.addfinalizer(teardown)

        return session

    def test_parameter_processing(self, session, Account):
        account = Account(
            balance=('USD', 15),
        )

        session.add(account)
        session.commit()

        account = session.query(Account).first()
        assert account.balance.currency == 'USD'
        assert account.balance.amount == 15



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import EmailType


@pytest.fixture
def User(Base):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, primary_key=True)
        email = sa.Column(EmailType)

        def __repr__(self):
            return 'Building(%r)' % self.id
    return User


@pytest.fixture
def init_models(User):
    pass


class TestCaseInsensitiveComparator(object):

    def test_supports_equals(self, session, User):
        query = (
            session.query(User)
            .filter(User.email == u'email@example.com')
        )

        assert 'user.email = lower(?)' in str(query)

    def test_supports_in_(self, session, User):
        query = (
            session.query(User)
            .filter(User.email.in_([u'email@example.com', u'a']))
        )
        assert (
            'user.email IN (lower(?), lower(?))'
            in str(query)
        )

    def test_supports_notin_(self, session, User):
        query = (
            session.query(User)
            .filter(User.email.notin_([u'email@example.com', u'a']))
        )
        assert (
            'user.email NOT IN (lower(?), lower(?))'
            in str(query)
        )

    def test_does_not_apply_lower_to_types_that_are_already_lowercased(
        self,
        User
    ):
        assert str(User.email == User.email) == (
            '"user".email = "user".email'
        )



import pytest
import sqlalchemy as sa

from sqlalchemy_utils.observer import observes


@pytest.fixture
def Catalog(Base):
    class Catalog(Base):
        __tablename__ = 'catalog'
        id = sa.Column(sa.Integer, primary_key=True)
        product_count = sa.Column(sa.Integer, default=0)

        @observes('categories.sub_category.products')
        def product_observer(self, products):
            self.product_count = len(products)

        categories = sa.orm.relationship('Category', backref='catalog')
    return Catalog


@pytest.fixture
def Category(Base):
    class Category(Base):
        __tablename__ = 'category'
        id = sa.Column(sa.Integer, primary_key=True)
        catalog_id = sa.Column(sa.Integer, sa.ForeignKey('catalog.id'))

        sub_category = sa.orm.relationship(
            'SubCategory',
            uselist=False,
            backref='category'
        )
    return Category


@pytest.fixture
def SubCategory(Base):
    class SubCategory(Base):
        __tablename__ = 'sub_category'
        id = sa.Column(sa.Integer, primary_key=True)
        category_id = sa.Column(sa.Integer, sa.ForeignKey('category.id'))
        products = sa.orm.relationship('Product', backref='sub_category')
    return SubCategory


@pytest.fixture
def Product(Base):
    class Product(Base):
        __tablename__ = 'product'
        id = sa.Column(sa.Integer, primary_key=True)
        price = sa.Column(sa.Numeric)

        sub_category_id = sa.Column(
            sa.Integer, sa.ForeignKey('sub_category.id')
        )
    return Product


@pytest.fixture
def init_models(Catalog, Category, SubCategory, Product):
    pass


@pytest.fixture
def catalog(session, Catalog, Category, SubCategory, Product):
    sub_category = SubCategory(products=[Product()])
    category = Category(sub_category=sub_category)
    catalog = Catalog(categories=[category])
    session.add(catalog)
    session.flush()
    return catalog


@pytest.mark.usefixtures('postgresql_dsn')
class TestObservesForOneToManyToOneToMany(object):

    def test_simple_insert(self, catalog):
        assert catalog.product_count == 1

    def test_add_leaf_object(self, catalog, session, Product):
        product = Product()
        catalog.categories[0].sub_category.products.append(product)
        session.flush()
        assert catalog.product_count == 2

    def test_remove_leaf_object(self, catalog, session, Product):
        product = Product()
        catalog.categories[0].sub_category.products.append(product)
        session.flush()
        session.delete(product)
        session.flush()
        assert catalog.product_count == 1

    def test_delete_intermediate_object(self, catalog, session):
        session.delete(catalog.categories[0].sub_category)
        session.commit()
        assert catalog.product_count == 0

    def test_gathered_objects_are_distinct(
        self,
        session,
        Catalog,
        Category,
        SubCategory,
        Product
    ):
        catalog = Catalog()
        category = Category(catalog=catalog)
        product = Product()
        category.sub_category = SubCategory(products=[product])
        session.add(
            Category(catalog=catalog, sub_category=category.sub_category)
        )
        session.commit()
        assert catalog.product_count == 1



import pytest
import sqlalchemy as sa

from sqlalchemy_utils.observer import observes


@pytest.fixture
def Catalog(Base):
    class Catalog(Base):
        __tablename__ = 'catalog'
        id = sa.Column(sa.Integer, primary_key=True)
        product_price = sa.Column(sa.Integer)

        @observes('category.sub_category.product')
        def product_observer(self, product):
            self.product_price = product.price if product else None

        category = sa.orm.relationship(
            'Category',
            uselist=False,
            backref='catalog'
        )
    return Catalog


@pytest.fixture
def Category(Base):
    class Category(Base):
        __tablename__ = 'category'
        id = sa.Column(sa.Integer, primary_key=True)
        catalog_id = sa.Column(sa.Integer, sa.ForeignKey('catalog.id'))

        sub_category = sa.orm.relationship(
            'SubCategory',
            uselist=False,
            backref='category'
        )
    return Category


@pytest.fixture
def SubCategory(Base):
    class SubCategory(Base):
        __tablename__ = 'sub_category'
        id = sa.Column(sa.Integer, primary_key=True)
        category_id = sa.Column(sa.Integer, sa.ForeignKey('category.id'))
        product = sa.orm.relationship(
            'Product',
            uselist=False,
            backref='sub_category'
        )
    return SubCategory


@pytest.fixture
def Product(Base):
    class Product(Base):
        __tablename__ = 'product'
        id = sa.Column(sa.Integer, primary_key=True)
        price = sa.Column(sa.Integer)

        sub_category_id = sa.Column(
            sa.Integer, sa.ForeignKey('sub_category.id')
        )
    return Product


@pytest.fixture
def init_models(Catalog, Category, SubCategory, Product):
    pass


@pytest.fixture
def catalog(session, Catalog, Category, SubCategory, Product):
    sub_category = SubCategory(product=Product(price=123))
    category = Category(sub_category=sub_category)
    catalog = Catalog(category=category)
    session.add(catalog)
    session.flush()
    return catalog


@pytest.mark.usefixtures('postgresql_dsn')
class TestObservesForOneToOneToOneToOne(object):

    def test_simple_insert(self, catalog):
        assert catalog.product_price == 123

    def test_replace_leaf_object(self, catalog, session, Product):
        product = Product(price=44)
        catalog.category.sub_category.product = product
        session.flush()
        assert catalog.product_price == 44

    def test_delete_leaf_object(self, catalog, session):
        session.delete(catalog.category.sub_category.product)
        session.flush()
        assert catalog.product_price is None



import pytest
import sqlalchemy as sa
from sqlalchemy.ext.hybrid import hybrid_property

from sqlalchemy_utils.relationships import select_correlated_expression


@pytest.fixture
def group_user_tbl(Base):
    return sa.Table(
        'group_user',
        Base.metadata,
        sa.Column('user_id', sa.Integer, sa.ForeignKey('user.id')),
        sa.Column('group_id', sa.Integer, sa.ForeignKey('group.id'))
    )


@pytest.fixture
def group_tbl(Base):
    class Group(Base):
        __tablename__ = 'group'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.String)

    return Group


@pytest.fixture
def friendship_tbl(Base):
    return sa.Table(
        'friendships',
        Base.metadata,
        sa.Column(
            'friend_a_id',
            sa.Integer,
            sa.ForeignKey('user.id'),
            primary_key=True
        ),
        sa.Column(
            'friend_b_id',
            sa.Integer,
            sa.ForeignKey('user.id'),
            primary_key=True
        )
    )


@pytest.fixture
def User(Base, group_user_tbl, friendship_tbl):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.String)
        groups = sa.orm.relationship(
            'Group',
            secondary=group_user_tbl,
            backref='users'
        )

        # this relationship is used for persistence
        friends = sa.orm.relationship(
            'User',
            secondary=friendship_tbl,
            primaryjoin=id == friendship_tbl.c.friend_a_id,
            secondaryjoin=id == friendship_tbl.c.friend_b_id,
        )

    friendship_union = (
        sa.select([
            friendship_tbl.c.friend_a_id,
            friendship_tbl.c.friend_b_id
        ]).union(
            sa.select([
                friendship_tbl.c.friend_b_id,
                friendship_tbl.c.friend_a_id]
            )
        ).alias()
    )

    User.all_friends = sa.orm.relationship(
        'User',
        secondary=friendship_union,
        primaryjoin=User.id == friendship_union.c.friend_a_id,
        secondaryjoin=User.id == friendship_union.c.friend_b_id,
        viewonly=True,
        order_by=User.id
    )
    return User


@pytest.fixture
def Category(Base, group_user_tbl, friendship_tbl):
    class Category(Base):
        __tablename__ = 'category'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.String)
        created_at = sa.Column(sa.DateTime)
        parent_id = sa.Column(sa.Integer, sa.ForeignKey('category.id'))
        parent = sa.orm.relationship(
            'Category',
            backref='subcategories',
            remote_side=[id],
            order_by=id
        )
    return Category


@pytest.fixture
def Article(Base, Category, User):
    class Article(Base):
        __tablename__ = 'article'
        id = sa.Column('_id', sa.Integer, primary_key=True)
        name = sa.Column(sa.String)
        name_synonym = sa.orm.synonym('name')

        @hybrid_property
        def name_upper(self):
            return self.name.upper() if self.name else None

        @name_upper.expression
        def name_upper(cls):
            return sa.func.upper(cls.name)

        content = sa.Column(sa.String)

        category_id = sa.Column(sa.Integer, sa.ForeignKey(Category.id))
        category = sa.orm.relationship(Category, backref='articles')

        author_id = sa.Column(sa.Integer, sa.ForeignKey(User.id))
        author = sa.orm.relationship(
            User,
            primaryjoin=author_id == User.id,
            backref='authored_articles'
        )

        owner_id = sa.Column(sa.Integer, sa.ForeignKey(User.id))
        owner = sa.orm.relationship(
            User,
            primaryjoin=owner_id == User.id,
            backref='owned_articles'
        )
    return Article


@pytest.fixture
def Comment(Base, Article, User):
    class Comment(Base):
        __tablename__ = 'comment'
        id = sa.Column(sa.Integer, primary_key=True)
        content = sa.Column(sa.String)
        article_id = sa.Column(sa.Integer, sa.ForeignKey(Article.id))
        article = sa.orm.relationship(Article, backref='comments')

        author_id = sa.Column(sa.Integer, sa.ForeignKey(User.id))
        author = sa.orm.relationship(User, backref='comments')

    Article.comment_count = sa.orm.column_property(
        sa.select([sa.func.count(Comment.id)])
        .where(Comment.article_id == Article.id)
        .correlate_except(Article)
    )

    return Comment


@pytest.fixture
def model_mapping(Article, Category, Comment, group_tbl, User):
    return {
        'articles': Article,
        'categories': Category,
        'comments': Comment,
        'groups': group_tbl,
        'users': User
    }


@pytest.fixture
def init_models(Article, Category, Comment, group_tbl, User):
    pass


@pytest.fixture
def dataset(
    session,
    User,
    group_tbl,
    Article,
    Category,
    Comment
):
    group = group_tbl(name='Group 1')
    group2 = group_tbl(name='Group 2')
    user = User(id=1, name='User 1', groups=[group, group2])
    user2 = User(id=2, name='User 2')
    user3 = User(id=3, name='User 3', groups=[group])
    user4 = User(id=4, name='User 4', groups=[group2])
    user5 = User(id=5, name='User 5')

    user.friends = [user2]
    user2.friends = [user3, user4]
    user3.friends = [user5]

    article = Article(
        name='Some article',
        author=user,
        owner=user2,
        category=Category(
            id=1,
            name='Some category',
            subcategories=[
                Category(
                    id=2,
                    name='Subcategory 1',
                    subcategories=[
                        Category(
                            id=3,
                            name='Subsubcategory 1',
                            subcategories=[
                                Category(
                                    id=5,
                                    name='Subsubsubcategory 1',
                                ),
                                Category(
                                    id=6,
                                    name='Subsubsubcategory 2',
                                )
                            ]
                        )
                    ]
                ),
                Category(id=4, name='Subcategory 2'),
            ]
        ),
        comments=[
            Comment(
                content='Some comment',
                author=user
            )
        ]
    )
    session.add(user3)
    session.add(user4)
    session.add(article)
    session.commit()


@pytest.mark.usefixtures('dataset', 'postgresql_dsn')
class TestSelectCorrelatedExpression(object):
    @pytest.mark.parametrize(
        ('model_key', 'related_model_key', 'path', 'result'),
        (
            (
                'categories',
                'categories',
                'subcategories',
                [
                    (1, 2),
                    (2, 1),
                    (3, 2),
                    (4, 0),
                    (5, 0),
                    (6, 0)
                ]
            ),
            (
                'articles',
                'comments',
                'comments',
                [
                    (1, 1),
                ]
            ),
            (
                'users',
                'groups',
                'groups',
                [
                    (1, 2),
                    (2, 0),
                    (3, 1),
                    (4, 1),
                    (5, 0)
                ]
            ),
            (
                'users',
                'users',
                'all_friends',
                [
                    (1, 1),
                    (2, 3),
                    (3, 2),
                    (4, 1),
                    (5, 1)
                ]
            ),
            (
                'users',
                'users',
                'all_friends.all_friends',
                [
                    (1, 3),
                    (2, 2),
                    (3, 3),
                    (4, 3),
                    (5, 2)
                ]
            ),
            (
                'users',
                'users',
                'groups.users',
                [
                    (1, 3),
                    (2, 0),
                    (3, 2),
                    (4, 2),
                    (5, 0)
                ]
            ),
            (
                'groups',
                'articles',
                'users.authored_articles',
                [
                    (1, 1),
                    (2, 1),
                ]
            ),
            (
                'categories',
                'categories',
                'subcategories.subcategories',
                [
                    (1, 1),
                    (2, 2),
                    (3, 0),
                    (4, 0),
                    (5, 0),
                    (6, 0)
                ]
            ),
            (
                'categories',
                'categories',
                'subcategories.subcategories.subcategories',
                [
                    (1, 2),
                    (2, 0),
                    (3, 0),
                    (4, 0),
                    (5, 0),
                    (6, 0)
                ]
            ),
        )
    )
    def test_returns_correct_results(
        self,
        session,
        model_mapping,
        model_key,
        related_model_key,
        path,
        result
    ):
        model = model_mapping[model_key]
        alias = sa.orm.aliased(model_mapping[related_model_key])
        aggregate = select_correlated_expression(
            model,
            sa.func.count(sa.distinct(alias.id)),
            path,
            alias
        )

        query = session.query(
            model.id,
            aggregate.label('count')
        ).order_by(model.id)
        assert query.all() == result

    def test_order_by_intermediate_table_column(
        self,
        session,
        model_mapping,
        group_user_tbl
    ):
        model = model_mapping['users']
        alias = sa.orm.aliased(model_mapping['groups'])
        aggregate = select_correlated_expression(
            model,
            sa.func.json_build_object('id', alias.id),
            'groups',
            alias,
            order_by=[group_user_tbl.c.user_id]
        ).alias('test')
        # Just check that the query execution doesn't fail because of wrongly
        # constructed aliases
        assert session.execute(aggregate)

    def test_with_non_aggregate_function(
        self,
        session,
        User,
        Article
    ):
        aggregate = select_correlated_expression(
            Article,
            sa.func.json_build_object('name', User.name),
            'comments.author',
            User
        )

        query = session.query(
            Article.id,
            aggregate.label('author_json')
        ).order_by(Article.id)
        result = query.all()
        assert result == [
            (1, {'name': 'User 1'})
        ]



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import generic_relationship

from . import GenericRelationshipTestCase


@pytest.fixture
def Building(Base):
    class Building(Base):
        __tablename__ = 'building'
        id = sa.Column(sa.Integer, primary_key=True)
    return Building


@pytest.fixture
def User(Base):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(sa.Integer, primary_key=True)
    return User


@pytest.fixture
def Event(Base):
    class Event(Base):
        __tablename__ = 'event'
        id = sa.Column(sa.Integer, primary_key=True)

        object_type = sa.Column(sa.Unicode(255), name="objectType")
        object_id = sa.Column(sa.Integer, nullable=False)

        object = generic_relationship(object_type, object_id)
    return Event


@pytest.fixture
def init_models(Building, User, Event):
    pass


class TestGenericRelationship(GenericRelationshipTestCase):
    pass



import pytest
import sqlalchemy as sa

from sqlalchemy_utils.observer import observes


@pytest.fixture
def Device(Base):
    class Device(Base):
        __tablename__ = 'device'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column(sa.String)
    return Device


@pytest.fixture
def Order(Base):
    class Order(Base):
        __tablename__ = 'order'
        id = sa.Column(sa.Integer, primary_key=True)

        device_id = sa.Column(
            'device', sa.ForeignKey('device.id'), nullable=False
        )
        device = sa.orm.relationship('Device', backref='orders')
    return Order


@pytest.fixture
def SalesInvoice(Base):
    class SalesInvoice(Base):
        __tablename__ = 'sales_invoice'
        id = sa.Column(sa.Integer, primary_key=True)
        order_id = sa.Column(
            'order',
            sa.ForeignKey('order.id'),
            nullable=False
        )
        order = sa.orm.relationship(
            'Order',
            backref=sa.orm.backref(
                'invoice',
                uselist=False
            )
        )
        device_name = sa.Column(sa.String)

        @observes('order.device')
        def process_device(self, device):
            self.device_name = device.name

    return SalesInvoice


@pytest.fixture
def init_models(Device, Order, SalesInvoice):
    pass


@pytest.mark.usefixtures('postgresql_dsn')
class TestObservesForOneToManyToOneToMany(object):

    def test_observable_root_obj_is_none(self, session, Device, Order):
        order = Order(device=Device(name='Something'))
        session.add(order)
        session.flush()



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import sort_query
from sqlalchemy_utils.functions import QuerySorterException

from . import assert_contains


class TestSortQuery(object):
    def test_without_sort_param_returns_the_query_object_untouched(
        self,
        session,
        Article
    ):
        query = session.query(Article)
        query = sort_query(query, '')
        assert query == query

    def test_column_ascending(self, session, Article):
        query = sort_query(session.query(Article), 'name')
        assert_contains('ORDER BY article.name ASC', query)

    def test_column_descending(self, session, Article):
        query = sort_query(session.query(Article), '-name')
        assert_contains('ORDER BY article.name DESC', query)

    def test_skips_unknown_columns(self, session, Article):
        query = session.query(Article)
        query = sort_query(query, '-unknown')
        assert query == query

    def test_non_silent_mode(self, session, Article):
        query = session.query(Article)
        with pytest.raises(QuerySorterException):
            sort_query(query, '-unknown', silent=False)

    def test_join(self, session, Article):
        query = (
            session.query(Article)
            .join(Article.category)
        )
        query = sort_query(query, 'name', silent=False)
        assert_contains('ORDER BY article.name ASC', query)

    def test_calculated_value_ascending(self, session, Article, Category):
        query = session.query(
            Category, sa.func.count(Article.id).label('articles')
        )
        query = sort_query(query, 'articles')
        assert_contains('ORDER BY articles ASC', query)

    def test_calculated_value_descending(self, session, Article, Category):
        query = session.query(
            Category, sa.func.count(Article.id).label('articles')
        )
        query = sort_query(query, '-articles')
        assert_contains('ORDER BY articles DESC', query)

    def test_subqueried_scalar(self, session, Article, Category):
        article_count = (
            sa.sql.select(
                [sa.func.count(Article.id)],
                from_obj=[Article.__table__]
            )
            .where(Article.category_id == Category.id)
            .correlate(Category.__table__)
        )

        query = session.query(
            Category, article_count.label('articles')
        )
        query = sort_query(query, '-articles')
        assert_contains('ORDER BY articles DESC', query)

    def test_aliased_joined_entity(self, session, Article, Category):
        alias = sa.orm.aliased(Category, name='categories')
        query = session.query(
            Article
        ).join(
            alias, Article.category
        )
        query = sort_query(query, '-categories-name')
        assert_contains('ORDER BY categories.name DESC', query)

    def test_joined_table_column(self, session, Article):
        query = session.query(Article).join(Article.category)
        query = sort_query(query, 'category-name')
        assert_contains('category.name ASC', query)

    def test_multiple_columns(self, session, Article):
        query = session.query(Article)
        query = sort_query(query, 'name', 'id')
        assert_contains('article.name ASC, article.id ASC', query)

    def test_column_property(self, session, Article, Category):
        Category.article_count = sa.orm.column_property(
            sa.select([sa.func.count(Article.id)])
            .where(Article.category_id == Category.id)
            .label('article_count')
        )

        query = session.query(Category)
        query = sort_query(query, 'article_count')
        assert_contains('article_count ASC', query)

    def test_column_property_descending(self, session, Article, Category):
        Category.article_count = sa.orm.column_property(
            sa.select([sa.func.count(Article.id)])
            .where(Article.category_id == Category.id)
            .label('article_count')
        )

        query = session.query(Category)
        query = sort_query(query, '-article_count')
        assert_contains('article_count DESC', query)

    def test_relationship_property(self, session, Category):
        query = session.query(Category)
        query = sort_query(query, 'articles')
        assert 'ORDER BY' not in str(query)

    def test_regular_property(self, session, Category):
        query = session.query(Category)
        query = sort_query(query, 'name_alias')
        assert 'ORDER BY' not in str(query)

    def test_synonym_property(self, session, Category):
        query = session.query(Category)
        query = sort_query(query, 'name_synonym')
        assert_contains('ORDER BY category.name ASC', query)

    def test_hybrid_property(self, session, Category):
        query = session.query(Category)
        query = sort_query(query, 'articles_count')
        assert_contains('ORDER BY (SELECT count(article.id) AS count_1', query)

    def test_hybrid_property_descending(self, session, Category):
        query = session.query(Category)
        query = sort_query(query, '-articles_count')
        assert_contains(
            'ORDER BY (SELECT count(article.id) AS count_1',
            query
        )
        assert ' DESC' in str(query)

    def test_assigned_hybrid_property(self, session, Article):
        def getter(self):
            return self.name

        Article.some_hybrid = sa.ext.hybrid.hybrid_property(
            fget=getter
        )
        query = session.query(Article)
        query = sort_query(query, 'some_hybrid')
        assert_contains('ORDER BY article.name ASC', query)

    def test_with_mapper_and_column_property(self, session, Base, Article):
        class Apple(Base):
            __tablename__ = 'apple'
            id = sa.Column(sa.Integer, primary_key=True)
            article_id = sa.Column(sa.Integer, sa.ForeignKey(Article.id))

        Article.apples = sa.orm.relationship(Apple)

        Article.apple_count = sa.orm.column_property(
            sa.select([sa.func.count(Apple.id)])
            .where(Apple.article_id == Article.id)
            .correlate(Article.__table__)
            .label('apple_count'),
            deferred=True
        )
        query = (
            session.query(sa.inspect(Article))
            .outerjoin(Article.apples)
            .options(
                sa.orm.undefer(Article.apple_count)
            )
            .options(sa.orm.contains_eager(Article.apples))
        )
        query = sort_query(query, 'apple_count')
        assert 'ORDER BY apple_count' in str(query)

    def test_table(self, session, Article):
        query = session.query(Article.__table__)
        query = sort_query(query, 'name')
        assert_contains('ORDER BY article.name', query)


@pytest.mark.usefixtures('postgresql_dsn')
class TestSortQueryRelationshipCounts(object):
    """
    Currently this doesn't work with SQLite
    """

    def test_relation_hybrid_property(self, session, Article):
        query = (
            session.query(Article)
            .join(Article.category)
        ).group_by(Article.id)
        query = sort_query(query, '-category-articles_count')
        assert_contains('ORDER BY (SELECT count(article.id) AS count_1', query)

    def test_aliased_hybrid_property(self, session, Article, Category):
        alias = sa.orm.aliased(
            Category,
            name='categories'
        )
        query = (
            session.query(Article)
            .outerjoin(alias, Article.category)
            .options(
                sa.orm.contains_eager(Article.category, alias=alias)
            )
        ).group_by(alias.id, Article.id)
        query = sort_query(query, '-categories-articles_count')
        assert_contains('ORDER BY (SELECT count(article.id) AS count_1', query)

    def test_aliased_concat_hybrid_property(self, session, Article, Category):
        alias = sa.orm.aliased(
            Category,
            name='aliased'
        )
        query = (
            session.query(Article)
            .outerjoin(alias, Article.category)
            .options(
                sa.orm.contains_eager(Article.category, alias=alias)
            )
        )
        query = sort_query(query, 'aliased-full_name')
        assert_contains(
            'concat(aliased.title, %(concat_1)s, aliased.name)', query
        )


@pytest.mark.usefixtures('postgresql_dsn')
class TestSortQueryWithPolymorphicInheritance(object):
    """
    Currently this doesn't work with SQLite
    """

    @pytest.fixture
    def TextItem(self, Base):
        class TextItem(Base):
            __tablename__ = 'text_item'
            id = sa.Column(sa.Integer, primary_key=True)

            type = sa.Column(sa.Unicode(255))

            __mapper_args__ = {
                'polymorphic_on': type,
                'with_polymorphic': '*'
            }
        return TextItem

    @pytest.fixture
    def Article(self, TextItem):
        class Article(TextItem):
            __tablename__ = 'article'
            id = sa.Column(
                sa.Integer, sa.ForeignKey(TextItem.id), primary_key=True
            )
            category = sa.Column(sa.Unicode(255))
            __mapper_args__ = {
                'polymorphic_identity': u'article'
            }
        return Article

    @pytest.fixture
    def init_models(self, TextItem, Article):
        pass

    def test_column_property(self, session, TextItem):
        TextItem.item_count = sa.orm.column_property(
            sa.select(
                [
                    sa.func.count('1')
                ],
            )
            .select_from(TextItem.__table__)
            .label('item_count')
        )

        query = sort_query(
            session.query(TextItem),
            'item_count'
        )
        assert_contains('ORDER BY item_count', query)

    def test_child_class_attribute(self, session, TextItem):
        query = sort_query(
            session.query(TextItem),
            'category'
        )
        assert_contains('ORDER BY article.category ASC', query)

    def test_with_ambiguous_column(self, session, TextItem):
        query = sort_query(
            session.query(TextItem),
            'id'
        )
        assert_contains('ORDER BY text_item.id ASC', query)


@pytest.mark.usefixtures('postgresql_dsn')
class TestSortQueryWithCustomPolymorphic(object):
    """
    Currently this doesn't work with SQLite
    """

    @pytest.fixture
    def TextItem(self, Base):
        class TextItem(Base):
            __tablename__ = 'text_item'
            id = sa.Column(sa.Integer, primary_key=True)

            type = sa.Column(sa.Unicode(255))

            __mapper_args__ = {
                'polymorphic_on': type,
            }
        return TextItem

    @pytest.fixture
    def Article(self, TextItem):
        class Article(TextItem):
            __tablename__ = 'article'
            id = sa.Column(
                sa.Integer, sa.ForeignKey(TextItem.id), primary_key=True
            )
            category = sa.Column(sa.Unicode(255))
            __mapper_args__ = {
                'polymorphic_identity': u'article'
            }
        return Article

    @pytest.fixture
    def BlogPost(self, TextItem):
        class BlogPost(TextItem):
            __tablename__ = 'blog_post'
            id = sa.Column(
                sa.Integer, sa.ForeignKey(TextItem.id), primary_key=True
            )
            __mapper_args__ = {
                'polymorphic_identity': u'blog_post'
            }
        return BlogPost

    def test_with_unknown_column(self, session, TextItem, BlogPost):
        query = sort_query(
            session.query(
                sa.orm.with_polymorphic(TextItem, [BlogPost])
            ),
            'category'
        )
        assert 'ORDER BY' not in str(query)

    def test_with_existing_column(self, session, TextItem, Article):
        query = sort_query(
            session.query(
                sa.orm.with_polymorphic(TextItem, [Article])
            ),
            'category'
        )
        assert 'ORDER BY' in str(query)



import pytest

from sqlalchemy_utils import get_bind


class TestGetBind(object):
    def test_with_session(self, session, connection):
        assert get_bind(session) == connection

    def test_with_connection(self, session, connection):
        assert get_bind(connection) == connection

    def test_with_model_object(self, session, connection, Article):
        article = Article()
        session.add(article)
        assert get_bind(article) == connection

    def test_with_unknown_type(self):
        with pytest.raises(TypeError):
            get_bind(None)



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import get_columns


@pytest.fixture
def Building(Base):
    class Building(Base):
        __tablename__ = 'building'
        id = sa.Column('_id', sa.Integer, primary_key=True)
        name = sa.Column('_name', sa.Unicode(255))
    return Building


class TestGetColumns(object):

    def test_table(self, Building):
        assert isinstance(
            get_columns(Building.__table__),
            sa.sql.base.ImmutableColumnCollection
        )

    def test_instrumented_attribute(self, Building):
        assert get_columns(Building.id) == [Building.__table__.c._id]

    def test_column_property(self, Building):
        assert get_columns(Building.id.property) == [
            Building.__table__.c._id
        ]

    def test_column(self, Building):
        assert get_columns(Building.__table__.c._id) == [
            Building.__table__.c._id
        ]

    def test_declarative_class(self, Building):
        assert isinstance(
            get_columns(Building),
            sa.util._collections.OrderedProperties
        )

    def test_declarative_object(self, Building):
        assert isinstance(
            get_columns(Building()),
            sa.util._collections.OrderedProperties
        )

    def test_mapper(self, Building):
        assert isinstance(
            get_columns(Building.__mapper__),
            sa.util._collections.OrderedProperties
        )

    def test_class_alias(self, Building):
        assert isinstance(
            get_columns(sa.orm.aliased(Building)),
            sa.util._collections.OrderedProperties
        )

    def test_table_alias(self, Building):
        alias = sa.orm.aliased(Building.__table__)
        assert isinstance(
            get_columns(alias),
            sa.sql.base.ImmutableColumnCollection
        )



from copy import copy

import pytest
import sqlalchemy as sa

from sqlalchemy_utils import get_column_key


@pytest.fixture
def Building(Base):
    class Building(Base):
        __tablename__ = 'building'
        id = sa.Column(sa.Integer, primary_key=True)
        name = sa.Column('_name', sa.Unicode(255))
    return Building


@pytest.fixture
def Movie(Base):
    class Movie(Base):
        __tablename__ = 'movie'
        id = sa.Column(sa.Integer, primary_key=True)
    return Movie


class TestGetColumnKey(object):

    def test_supports_aliases(self, Building):
        assert (
            get_column_key(Building, Building.__table__.c.id) ==
            'id'
        )
        assert (
            get_column_key(Building, Building.__table__.c._name) ==
            'name'
        )

    def test_supports_vague_matching_of_column_objects(self, Building):
        column = copy(Building.__table__.c._name)
        assert get_column_key(Building, column) == 'name'

    def test_throws_value_error_for_unknown_column(self, Building, Movie):
        with pytest.raises(sa.orm.exc.UnmappedColumnError):
            get_column_key(Building, Movie.__table__.c.id)



import sys
from datetime import datetime

import pytest
import sqlalchemy as sa

from sqlalchemy_utils import generic_repr, Timestamp


class TestTimestamp(object):
    @pytest.fixture
    def Article(self, Base):
        class Article(Base, Timestamp):
            __tablename__ = 'article'
            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.Unicode(255), default=u'Some article')
        return Article

    def test_created(self, session, Article):
        then = datetime.utcnow()
        article = Article()

        session.add(article)
        session.commit()

        assert article.created >= then and article.created <= datetime.utcnow()

    def test_updated(self, session, Article):
        article = Article()

        session.add(article)
        session.commit()

        then = datetime.utcnow()
        article.name = u"Something"

        session.commit()

        assert article.updated >= then and article.updated <= datetime.utcnow()


class TestGenericRepr:
    @pytest.fixture
    def Article(self, Base):
        class Article(Base):
            __tablename__ = 'article'
            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.Unicode(255), default=u'Some article')
        return Article

    def test_repr(self, Article):
        """Representation of a basic model."""
        Article = generic_repr(Article)
        article = Article(id=1, name=u'Foo')
        if sys.version_info[0] == 2:
            expected_repr = u'Article(id=1, name=u\'Foo\')'
        elif sys.version_info[0] == 3:
            expected_repr = u'Article(id=1, name=\'Foo\')'
        else:
            raise AssertionError
        actual_repr = repr(article)

        assert actual_repr == expected_repr

    def test_repr_partial(self, Article):
        """Representation of a basic model with selected fields."""
        Article = generic_repr('id')(Article)
        article = Article(id=1, name=u'Foo')
        expected_repr = u'Article(id=1)'
        actual_repr = repr(article)

        assert actual_repr == expected_repr

    def test_not_loaded(self, session, Article):
        """:py:func:`~sqlalchemy_utils.models.generic_repr` doesn't force
        execution of additional queries if some fields are not loaded and
        instead represents them as "<not loaded>".
        """
        Article = generic_repr(Article)
        article = Article(name=u'Foo')
        session.add(article)
        session.commit()

        article = session.query(Article).options(sa.orm.defer('name')).one()
        actual_repr = repr(article)

        expected_repr = u'Article(id={}, name=<not loaded>)'.format(article.id)
        assert actual_repr == expected_repr
